Rice.require 'TableConverter'
Rice.require 'ExtensionLoader'
Rice.require 'commandString'
Rice.require 'Native'
Rice.require 'UTF8'


-----------------------------------------------------------------------------
-- Local TABLE : INPUTTRAY COMMAND
-----------------------------------------------------------------------------
local INPUTTRAY_TABLE =	{
	autotrayselect		= 'ALL',
	tray1				= 'TRAY1',
	tray2				= 'TRAY2',
	tray3				= 'TRAY3',
	tray3lct			= 'LCT',
	tray4				= 'TRAY4',
	tray5				= 'TRAY5',
	tray6				= 'TRAY6',
	tray7				= 'TRAY7',
	tray8				= 'TRAY8',
	tray9				= 'TRAY9',
	tray10				= 'TRAY10',
	tray11				= 'TRAY11',
	traya				= 'TRAYA',
	bypasstray			= 'BYPASS',
	onesheetbypass		= 'TRAY7',
	frontbypasstray		= 'BYPASS',
	largecapacitytray	= 'LCT',
	interposer			= 'INSERTER',
	interposeruppertray	= 'INSERTER',
	interposerlowertray	= 'INSERTER2',
	bindinginterposer1	= 'BINDINSERTER',
	bindinginterposer2	= 'BINDINSERTER2',
	largecapacityinterposer1	= 'LARGECAPACITYINSERTER',
	largecapacityinterposer2	= 'LARGECAPACITYINSERTER2'
}

-----------------------------------------------------------------------------
-- Local TABLE : OUTPUTTRAY COMMAND
-----------------------------------------------------------------------------
local OUTPUTTRAY_TABLE =	{
	printerdefault			= nil,
	standardtray			= 'UPPER',
	copytray				= 'UPPER',
	internalshifttray		= 'UPPER',
	shifttray				= 'UPPER',
	internaltray			= 'UPPER',
	internaltray2			= 'INNER',
	finisheruppertray		= 'FINISHERPROOF',
	finishershifttray		= 'FINISHERSHIFT',
	finishershifttray1		= 'FINISHERPROOF',
	finishershifttray2		= 'FINISHERSHIFT',
	finishershifttraysecond	= 'FINISHERSHIFT2',
	bookletprocessortray	= 'FINISHERSHIFT',
	bookletprocessortray2	= 'FINISHERBOOKLET',
	finisherbooklettray		= 'FINISHERBOOKLET',
	trimmertray				= 'FINISHERBOOKLET',
	lefttray				= 'LOWER',
	lefttray1				= 'LOWER2',
	lefttray2				= 'LOWER3',
	righttray				= 'LOWER',
	mailboxprooftray		= 'OPTIONALOUTPUTBIN1',
	mailboxtray1			= 'OPTIONALOUTPUTBIN2',
	mailboxtray2			= 'OPTIONALOUTPUTBIN3',
	mailboxtray3			= 'OPTIONALOUTPUTBIN4',
	mailboxtray4			= 'OPTIONALOUTPUTBIN5',
	mailboxtray5			= 'OPTIONALOUTPUTBIN6',
	mailboxtray6			= 'OPTIONALOUTPUTBIN7',
	mailboxtray7			= 'OPTIONALOUTPUTBIN8',
	mailboxtray8			= 'OPTIONALOUTPUTBIN9',
	mailboxtray9			= 'OPTIONALOUTPUTBIN10',
	onesheetoutputtray		= 'LOWER',
	loadingtray				= 'LOADINGTRAY',
	ringstacker				= 'RINGSTACKER',
	largestackertray		= 'STACKER',
	largestackertray2		= 'STACKER2',
	longstackertray			= 'FINISHERSHIFT',
	longstackertray2		= 'FINISHERSHIFT2',
	stackerprooftray		= 'STACKERPROOF',
	stackerprooftray2		= 'STACKER2PROOF',
	foldprooftray			= 'FOLDPROOF',
	peripheralsifbox		= 'PERIPHERALIFBOX',
	twinloopbindtray        = 'TWINLOOPBIND'
}

-----------------------------------------------------------------------------
-- Local TABLE : STAPLE_TABLE COMMAND
-----------------------------------------------------------------------------
local STAPLE_TABLE = {
	off							= 'OFF',
	topleftslant				= 'LEFTTOPSLANTPORT',
	toprightslant				= 'RIGHTTOPSLANTPORT',
	topleft						= 'LEFTTOP',
	topright					= 'RIGHTTOP',
	['2attop']					= 'TOP2PORT',
	['2atleft']					= 'LEFT2PORT',
	['2atright']				= 'RIGHT2PORT',
	['2atcenter']				= 'BOOKLET',
	['topleftslantstapleless']	= 'STAPLELESSLEFTTOPSLANTPORT',
	['toprightslantstapleless']	= 'STAPLELESSRIGHTTOPSLANTPORT'
}

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isMember( value, tbl )
	for i = 1 , #tbl do
		if(tbl[i] == value) then
			return true
		end
	end
	return false
end

local function isEmpty( value )
	return ( not value or '' == value )
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function RoundOff( num, power_of_ten)
	if( not tonumber(num) or not tonumber(power_of_ten) ) then
		return nil
	end

	return tostring((math.floor(tonumber( num ) * (10 ^ ( -1 * tonumber(power_of_ten))) + 0.5)) * (10 ^ tonumber(power_of_ten)))
end

local function Multiply( a, b )
	if( not tonumber(a) or not tonumber(b) ) then
		return nil
	end

	return tostring(a * b)
end

local function GetMMValue( num )
	local MMValue = tonumber(RoundOff( Multiply( num, 0.1 ) , -1))
	if( not MMValue ) then
		return nil
	else
		return string.format( '%0.1f', MMValue )
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isOrientationChanged(tbl)
	
	local result = false
	
	if( isMember( tbl.layout, { '2pagespersheet', '6pagespersheet' } ) ) then
		result = not result
	end
	
	if( isMember( tbl.booklet, { 'booklet' } ) ) then
		result = not result
	end
	
	if( '1pageinto2sheets' == tbl.poster ) then
		result = not result
	end
	
	return result
end

function isPortrait(tbl)
	if( isOrientationChanged(tbl) ) then
		return ( 'landscape' == tbl.orientation )
	else
		return ( 'landscape' ~= tbl.orientation )
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isCollate(tbl)
	if( 'off' == tbl.enablecollate ) then
		return false
	elseif( 'driver' == tbl.enablecollate ) then
		if( 'application' == tbl.collate ) then
			return ( '0' ~= tbl.dmCollate )
		elseif( 'driver' == tbl.collate ) then
			return true
		else
			return false
		end
	else
		if( 'application' == tbl.collate ) then
			return ( '0' ~= tbl.dmCollate )
		elseif( 'driver' == tbl.collate ) then
			return false
		else
			return true
		end
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function getPrintSize( tbl )
	if( tbl ) then
		local documentsize = tbl.documentsize and tbl.documentsize or 'a4'
		local printsize = tbl.printsize and tbl.printsize or 'sameasoriginalsize'
		
		if( 'sameasoriginalsize' == printsize ) then
			return documentsize
		else
			return printsize
		end
	else
		return 'a4'
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function ExpandTable( t )
	local retTable = {}

	local function Expand( t )
		if( 'table' == type(t) ) then
			for i, it in pairs( t ) do
				if( 'table' == type( it ) ) then
					Expand( it )
				elseif( 'string' == type( it ) ) then
					table.insert( retTable, it )
				elseif( 'function' == type( it ) ) then
					Expand( it() )
				end
			end
		else
			if( 'string' == type( t ) ) then
				table.insert( retTable, t )
			elseif( 'function' == type( t ) ) then
				Expand( t() )
			end
		end
	end

	Expand( t )
	return retTable
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function GetEncryptPassword( tbl )
	if( isMember( tbl.jobtype, { 'documentserver', 'documentserverrgb' } ) ) then
		if('3' == tbl.jobpasswordgeneration ) then
			return tbl.jh_documentserver_password3 and tbl.jh_documentserver_password3 or ''
		elseif('4' == tbl.jobpasswordgeneration ) then
			return tbl.jh_documentserver_password4 and tbl.jh_documentserver_password4 or ''
		elseif('5' == tbl.jobpasswordgeneration ) then
			return tbl.jh_documentserver_password5 and tbl.jh_documentserver_password5 or ''
		else
			return tbl.jh_documentserver_password2 and tbl.jh_documentserver_password2 or ''
		end
	else
		if('3' == tbl.jobpasswordgeneration ) then
			return tbl.jh_jobpassword3 and tbl.jh_jobpassword3 or ''
		elseif('4' == tbl.jobpasswordgeneration ) then
			return tbl.jh_jobpassword4 and tbl.jh_jobpassword4 or ''
		elseif('5' == tbl.jobpasswordgeneration ) then
			return tbl.jh_jobpassword5 and tbl.jh_jobpassword5 or ''
		else
			return tbl.jh_jobpassword2 and tbl.jh_jobpassword2 or ''
		end
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isWrongString( str, max_length )
	if (not str) or (not max_length) then
		return true
	end

	if string.len(str) > max_length then
		return true
	else
		return false
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isFrontCoverWithBooklet( tbl )
	return ( 'on' == tbl.frontcover and '2atcenter' == tbl.staple and isMember( tbl.centerstaple_swapcover, {'1', '2', '3'} ) )
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function GetCommandCharcode( tbl )
	if ('on' == tbl.utf8mode) then
		return 'utf8'
	elseif( 'on' == tbl.gb18030support	and
		'8' == tbl.jh_charset			) then
		return 'utf8'
	elseif ('10' == tbl.jh_charset or '11' == tbl.jh_charset or '33' == tbl.jh_charset) and ('on' == tbl.output_utf8_rtl_language) then
		return 'utf8'
	else
		return 'osdependent'
	end
end

local function GetJobName( tbl )
	if( isEmpty(tbl.jobname) ) then
		return '""'
	else
		return '"' .. get_trimmed_command_string( tbl.jobname, GetCommandCharcode(tbl), 232 ) .. '"'
	end
end

local function GetJob_Name( tbl )
	if( isEmpty(tbl.jobname) ) then
		return '""'
	else
		return '"' .. get_trimmed_command_string( tbl.jobname, GetCommandCharcode(tbl), 80 ) .. '"'
	end
end

-----------------------------------------------------------------------------
-- Local Function :
-----------------------------------------------------------------------------
local function isReverseSetting( tbl )
	if( tbl.reverseorder				and
		'on'	== tbl.reverseorder		) then
		return true
	elseif(	 (('1' == tbl.centerstaple_reverseorder ) and ( '2atcenter' == tbl.staple ))) then
		return true
	elseif(( 'on' == tbl.multisheetfold ) and (false ~= isMember(tbl.fold, {'outsidehalffold', 'rightortopoutsidespiralfold', 'leftorbottomoutsidespiralfold'}) )) then
		return true
	elseif('superimposedfoldinside' == tbl.fold) then
		return true
	else
		return false	
	end
end

-----------------------------------------------------------------------------
-- Conversion Table : setting_table -> job_command_table
-----------------------------------------------------------------------------
local convert_table = {

	userid = {
		item = 'USERID',
		values = function(value, tbl)
			if( isEmpty( value ) or isWrongString( value, 233 ) ) then
				return '""'
			else
				return  '"' .. value .. '"'
			end
		end
	},

	jobtype = {
		item = function(value, tbl)
			local password = GetEncryptPassword( tbl )

			local function PasswordItem()
				if( isEmpty( password ) or ( '4' == tbl.jobpasswordgeneration and 'on' == tbl.userauthentication ) ) then
					return nil
				end

				if( '4' == tbl.jobpasswordgeneration ) then
					return {'JOBPASSWORD' .. tbl.jobpasswordgeneration, 'JOBPASSWORDENCRYPTMETHOD', 'JOBPASSWORDENCRYPTSETTING'}
				elseif( true == isMember( tbl.jobpasswordgeneration, { '2', '3', '5' } ) ) then
					return 'JOBPASSWORD' .. tbl.jobpasswordgeneration
				else
					return 'JOBPASSWORD'
				end
			end

			local function JobIDItem()
				local fname = (function ()
					if( isMember( value, { 'documentserver', 'documentserverrgb' } ) ) then
						return get_trimmed_command_string( tbl.documentserver_filename, GetCommandCharcode(tbl), 20 )
					else
						return get_trimmed_command_string( tbl.filename, GetCommandCharcode(tbl), 20 )
					end
				end)()

				if ( isEmpty( fname ) ) then
					return nil
				else
					return 'JOBID'
				end
			end

			local function OwnerIDItem()
				if ( isEmpty( tbl.username ) ) then
					return nil
				else
					return 'OWNERID'
				end
			end

			local function ScheduledPrintItem()
				local hour		= tonumber( tbl.scheduledprint_hour )
				local minute	= tonumber( tbl.scheduledprint_minute )

				if( isEmpty( tbl.scheduledprint ) or isEmpty( hour ) or isEmpty( minute ) or ( 'on' ~= tbl.scheduledprint ) ) then
					return nil
				end

				if( ( 0 <= hour and hour <= 23 ) and ( 0<= minute and minute <= 59 ) ) then
					return 'PRINTAPPOINTEDTIME'
				else
					return nil
				end
			end
			
			local function FolderNumberItem()
				if( nil == tbl.documentserver_foldernumber ) then
					return nil
				else
					return 'FOLDERNUMBER'
				end
			end
			
			local function FolderPasswordItem()
				if ( isEmpty( tbl.jh_documentserver_folderpassword5 ) ) then
					return nil
				else
					return 'FOLDERPASSWORD'
				end
			end

			local t = {
				normalprint			= nil,
				sampleprint			= nil,
				lockedprint			= { nil			, nil			, nil		, PasswordItem	, nil			, nil					},
				holdprint			= { 'SAVEMODE'	, nil			, JobIDItem	, nil			, nil			, ScheduledPrintItem	},
				storedprint			= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			, nil					},
				storedprintshared	= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			, nil					},
				storeandprint		= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			, nil					},
				storeandprintshared	= { 'SAVEMODE'	, 'ACCESSMODE'	, JobIDItem	, PasswordItem	, nil			, nil					},
				documentserver		= { 'DISKIMAGE'	, nil			, JobIDItem	, PasswordItem	, OwnerIDItem	, FolderNumberItem		, FolderPasswordItem	},
				documentserverrgb	= { nil			, nil			, JobIDItem	, PasswordItem	, OwnerIDItem	, nil					}
			}

			return t[value] and ExpandTable( t[value] ) or ExpandTable( t['normalprint'] )
		end,
		values = function( value, tbl )
			local password = GetEncryptPassword( tbl )

			local function JobIDValue()
				local fname = (function ()
					if( isMember( value, { 'documentserver', 'documentserverrgb' } ) ) then
						return get_trimmed_command_string( tbl.documentserver_filename, GetCommandCharcode(tbl), 20 )
					else
						return get_trimmed_command_string( tbl.filename, GetCommandCharcode(tbl), 20 )
					end
				end)()

				if ( isEmpty( fname ) ) then
					return nil
				else
					return '"' .. fname .. '"'
				end
			end

			local function OwnerIDValue()
				local uname = get_trimmed_command_string( tbl.username, GetCommandCharcode(tbl), 20 )
				if ( isEmpty( uname ) ) then
					return nil
				else
					return '"' .. uname .. '"'
				end
			end

			local function PasswordValue( )
				if( isEmpty( password ) or ( '4' == tbl.jobpasswordgeneration and 'on' == tbl.userauthentication ) ) then
					return nil
				end

				if( '5' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 227 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				elseif( '4' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 227 ) ) then
						return nil
					else
						return { '"' .. password .. '"', '0', '2' }
					end
				elseif( '3' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 227 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				elseif( '2' == tbl.jobpasswordgeneration ) then
					if( isWrongString( password, 8 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				else
					if( isWrongString( password, 4 ) ) then
						return nil
					else
						return '"' .. password .. '"'
					end
				end
			end

			local function ScheduledPrintValue()
				local hour		= tonumber( tbl.scheduledprint_hour )
				local minute	= tonumber( tbl.scheduledprint_minute )

				if( isEmpty( tbl.scheduledprint ) or isEmpty( hour ) or isEmpty( minute ) or ( 'on' ~= tbl.scheduledprint ) ) then
					return nil
				end

				if( ( 0 <= hour and hour <= 23 ) and ( 0<= minute and minute <= 59 ) ) then
					return '"' .. string.format( '%02d', hour ) .. ':' .. string.format( '%02d', minute ) .. '"'
				else
					return nil
				end
			end

			local function FolderNumberValue()
				if( nil ~= tbl.documentserver_foldernumber ) then
					if( '' == tbl.documentserver_foldernumber ) then
						return '0'
					else
						return tbl.documentserver_foldernumber
					end
				else
					return nil
				end
			end
			
			local function FolderPasswordValue()
				if( isEmpty( tbl.jh_documentserver_folderpassword5 ) or isWrongString( tbl.jh_documentserver_folderpassword5, 225 ) ) then
					return nil
				else
					return ( '"' .. tbl.jh_documentserver_folderpassword5 .. '"' )
				end
			end

			local t = {
				normalprint			= nil,
				sampleprint			= nil,
				lockedprint			= { nil		, nil		, nil			, PasswordValue	, nil			, nil					},
				holdprint			= { 'OFF'	, nil		, JobIDValue	, nil			, nil			, ScheduledPrintValue	},
				storedprint			= { 'ON'	, 'PRIVATE'	, JobIDValue	, PasswordValue	, nil			, nil					},
				storedprintshared	= { 'ON'	, 'PUBLIC'	, JobIDValue	, PasswordValue	, nil			, nil					},
				storeandprint		= { 'ON'	, 'PRIVATE'	, JobIDValue	, PasswordValue	, nil			, nil					},
				storeandprintshared	= { 'ON'	, 'PUBLIC'	, JobIDValue	, PasswordValue	, nil			, nil					},
				documentserver		= { 'ON'	, nil		, JobIDValue	, PasswordValue	, OwnerIDValue	, FolderNumberValue	, FolderPasswordValue	},
				documentserverrgb	= { nil		, nil		, JobIDValue	, PasswordValue	, OwnerIDValue	, nil					}
			}

			return t[value] and ExpandTable( t[value] ) or ExpandTable( t['normalprint'] )
		end
	},

	userid_type = {
	
		item = 'DOCPRINTMENUUSER',
		values = function(value, tbl)
			if( 'userdefined' == tbl.userid_type ) then
				return 'USERID'
			elseif( 'windowsloginname' == tbl.userid_type ) then
				return 'HOSTLOGINNAME'
			else
				return nil
			end
		end
	},
		
	faceup = {
		item	= {'PAPERFACE'},
		values	= function( value, tbl )
			if( not isMember( tbl.fold, {	'rightzfold',
											'leftzfold',
											'bottomzfold',
											'insidefold',
											'outsidefold',
											'superimposedfoldinside',
											'superimposedfoldoutside',
											'insidehalffold',
											'outsidehalffold',
											'rightortopinsidespiralfold',
											'leftorbottominsidespiralfold',
											'rightortopoutsidespiralfold',
											'leftorbottomoutsidespiralfold',
											'outsidezigzagfold',
											'insidezigzagfold',
											'rightortopinsidedoublefold',
											'leftorbottominsidedoublefold',
											'rightortopoutsidedoublefold',
											'leftorbottomoutsidedoublefold',
											'insidedoublegatefold',
											'outsidedoublegatefold'
										 } ) ) then
					if( 'on' == value ) then
						return 'UP'
					else
						return 'DOWN'
					end
			else
					return nil
			end
		end
	},

	fold = {
		item = {
			off								=  { 'FOLD' },
			rightzfold						=  { 'FOLD', 'PAPERFACE' },
			leftzfold						=  { 'FOLD', 'PAPERFACE' },
			bottomzfold						=  { 'FOLD', 'PAPERFACE' },
			insidehalffold					=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			outsidehalffold					=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			rightortopinsidespiralfold		=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			leftorbottominsidespiralfold	=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			rightortopoutsidespiralfold		=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			leftorbottomoutsidespiralfold	=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			outsidezigzagfold				=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			insidezigzagfold				=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			rightortopinsidedoublefold		=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			leftorbottominsidedoublefold	=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			rightortopoutsidedoublefold		=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			leftorbottomoutsidedoublefold	=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			insidedoublegatefold			=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			outsidedoublegatefold			=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			insidefold						=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			outsidefold						=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			superimposedfoldinside			=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' },
			superimposedfoldoutside			=  { 'FOLD', 'FOLDOPENSIDE', 'PAPERFACE' }
		},
		values = function( value, tbl )
			local is_portrait = isPortrait( tbl )

			local function TopOrRight()
				if( is_portrait ) then
					return 'TOP'
				else
					return 'RIGHT'
				end
			end

			local function BottomOrLeft()
				if( is_portrait ) then
					return 'BOTTOM'
				else
					return 'LEFT'
				end
			end

			local function DownOrUp()
				if( 'on' == tbl.multisheetfold and 'off' ~= tbl.duplex ) then
					return 'DOWN'
				else
					return 'UP'
				end
			end

			local t = {
				off								= { 'OFF'			},
				rightzfold						= { 'ZRIGHT'		, 'DOWN'	},
				leftzfold						= { 'ZLEFT'			, 'DOWN'	},
				bottomzfold						= { 'ZBOTTOM'		, 'DOWN'	},
				insidefold						= { 'INSIDE2'		, TopOrRight	, 'DOWN'	},
				outsidefold						= { 'OUTSIDE2'		, TopOrRight	, 'DOWN'	},
				superimposedfoldinside			= { 'SUPERIMPOSE2'	, TopOrRight	, 'DOWN'	},
				superimposedfoldoutside			= { 'SUPERIMPOSE2'	, TopOrRight	, 'DOWN'	},
				insidehalffold					= { 'TWO'			, TopOrRight	, 'DOWN'	},
				outsidehalffold					= { 'TWO'			, BottomOrLeft	, DownOrUp	},
				rightortopinsidespiralfold		= { 'INSIDE3'		, TopOrRight	, 'DOWN'	},
				leftorbottominsidespiralfold	= { 'INSIDE3'		, BottomOrLeft	, 'DOWN'	},
				rightortopoutsidespiralfold		= { 'INSIDE3'		, TopOrRight	, DownOrUp	},
				leftorbottomoutsidespiralfold	= { 'INSIDE3'		, BottomOrLeft	, DownOrUp	},
				outsidezigzagfold				= { 'OUTSIDE3'		, TopOrRight	, 'DOWN'	},
				insidezigzagfold				= { 'OUTSIDE3'		, BottomOrLeft	, 'DOWN'	},
				rightortopinsidedoublefold		= { 'SIMPLE4'		, TopOrRight	, 'DOWN'	},
				leftorbottominsidedoublefold	= { 'SIMPLE4'		, BottomOrLeft	, 'DOWN'	},
				rightortopoutsidedoublefold		= { 'SIMPLE4'		, TopOrRight	, 'UP'		},
				leftorbottomoutsidedoublefold	= { 'SIMPLE4'		, BottomOrLeft	, 'UP'		},
				insidedoublegatefold			= { 'GATEFOLD4'		, TopOrRight	, 'DOWN'	},
				outsidedoublegatefold			= { 'GATEFOLD4'		, BottomOrLeft	, 'UP'		}
			}

			return ExpandTable( t[value] )
		end,
		default = { 'FOLD', 'OFF' }
	},

	orientation = {
		item = 'ORIENTATION',
		values = function( value, tbl )
			if( isPortrait( tbl ) ) then
				return isMember( tbl.booklet, { 'booklet' } ) and 'LANDSCAPE' or 'PORTRAIT'
			else
				return isMember( tbl.booklet, {'booklet' } ) and 'PORTRAIT' or 'LANDSCAPE'
			end
		end
	},

	duplex = {
		item = { 'DUPLEX', 'BINDING' },
		values = function( value, tbl )
			if( tbl.booklet and 'off' ~= tbl.booklet ) then
				return nil
			else
				if isMember( value, { 'opentoleft', 'opentotop' } ) then
					-- dl
					if( isOrientationChanged(tbl) ) then
						if( 'opentoleft' == value ) then
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'SHORTEDGE'}
							else
								return {'ON', 'LONGEDGE'}
							end
						else
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'LONGEDGE'}
							else
								return {'ON', 'SHORTEDGE'}
							end
						end
					else
						if( 'opentoleft' == value ) then
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'LONGEDGE'}
							else
								return {'ON', 'SHORTEDGE'}
							end
						else
							if( 'portrait' == tbl.orientation ) then
								return {'ON', 'SHORTEDGE'}
							else
								return {'ON', 'LONGEDGE'}
							end
						end
					end
				else
					-- Vdl
					local t = {
						off					= { 'OFF' },
						longedgeleftortop	= { 'ON', 'LONGEDGE' },
						shortedgetoporleft	= { 'ON', 'SHORTEDGE' }
					}

					return t[value] and t[value] or t['off']
				end
			end
		end
	},

	resolution = {
		item = { 'RESOLUTION', 'BITSPERDOT' },
		values = {
			['300'] 		= { '300'	, '1' },
			['600'] 		= { '600'	, '1' },
			['1200x600']	= { '600'	, '2' },
			['600x1200']	= { '600'	, '2' },
			['2400x600']	= { '600'	, '2' },
			['9600x600']	= { '600'	, '4' },
			['1200'] 		= { '1200'	, '1' },
			['4800x1200'] 	= { '1200'	, '2' },

		},
		default = { { 'RESOLUTION', 'BITSPERDOT' }, {'600', '1'} }
	},

	copies = {
		item = function( value, tbl )
			if( isCollate(tbl) or isMember( tbl.staple, { 'topleftslant', 'toprightslant', 'topleft', '2atright', '3atleft', '4atleft', 'multiatleft', 'multiatright', 'multiattop' } ) ) then
				if( 'driver' == tbl.enablecollate ) then
					return nil
				else
					return {'COPIES', 'QTY' }
				end
			elseif( 'driver' == tbl.collate ) then
				return nil
			else
				return 'COPIES'
			end
		end,
		values = function( value, tbl )
			if( isCollate(tbl) or isMember( tbl.staple, { 'topleftslant', 'toprightslant', 'topleft', '2atright', '3atleft', '4atleft', 'multiatleft', 'multiatright', 'multiattop' } ) ) then
				if( 'driver' == tbl.enablecollate ) then
					return nil
				else
					return {'1', value }
				end
			else
				return value
			end
		end
	},

	printpapercount = {
		item = 'JOBPAGECNT',
		values = function( value, tbl )
			if( 'driver' == tbl.collate ) then
				return value
			else
				if( 'driver' == tbl.enablecollate		and		'0' ~= tbl.dmCollate ) then
					return value
				else
					return nil
				end
			end
		end
	},

	colorbw = {
		item		= { 'DATAMODE', 'RENDERMODE' },
		values	= {
			color			= { 'COLOR'		, 'COLOR'		},
			blackandwhite	= { 'GRAYSCALE'	, 'GRAYSCALE'	}
		},
		default = { { 'DATAMODE', 'RENDERMODE' }, { 'COLOR' , 'COLOR' } }
	},

	jh_standardpages_mediatype_jobcommand = {
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return nil
			else
				return 'MEDIATYPE'
			end
		end,
		values = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return nil
			else
				return value
			end
		end
	},

	jh_transparency_mediatype_jobcommand = {
		item = 'MEDIATYPE',
		values = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return value
			else
				return nil
			end
		end
	},

	booklet = {
		item = {'DUPLEX', 'BINDING'},
		values = function( value, tbl )
			if( 'off' == value ) then
				return nil
			else
				if(	nil ~= tonumber(tbl.jh_print_paperwidth_mm)										and
					nil ~= tonumber(tbl.jh_print_paperheight_mm)									and
					tonumber(tbl.jh_print_paperwidth_mm) > tonumber(tbl.jh_print_paperheight_mm) 	and
					not isMember( tbl.layout, { '2pagespersheet', '6pagespersheet'} )	)	then
					return { 'ON', 'LONGEDGE' }
				else
					return { 'ON', 'SHORTEDGE' }
				end
			end
		end
	},

	insertseparationsheet = {
		item = function( value, tbl )
			if( 'on' == value ) then
				return {'SEPARATIONSHEETMODE', 'SEPARATIONSHEETCYCLE'}
			else
				return 'SEPARATIONSHEETMODE'
			end
		end,
		values = function( value, tbl )
			if( 'on' == value ) then
				return { 'SORT', tbl.insertseparationsheet_percopies }
			else
				return 'OFF'
			end
		end
	},

	insertseparationsheet_inputtray = {
		item = function( value, tbl )
			if( 'on' == tbl.insertseparationsheet ) then
				return 'SEPARATIONSHEETTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	donotprintblankpages = {
		item = 'WHITEPAPERSUPPRESS',
		values = function( value, tbl )
			if( 'on' == value ) then
				if( tbl.duplex						and
					'off' 	~= tbl.duplex			and
					isReverseSetting(tbl)			) then
					return 'OFF'
				elseif(	tbl.booklet				and
						'booklet' == tbl.booklet) then
					return 'OFF'
				else
					return 'ON'
				end
			else
				return 'OFF'
			end
		end,
		default = { 'WHITEPAPERSUPPRESS', 'OFF' }
	},
	
	edgetoedgeprint = {
		item = 'EDGETOEDGE',
		values = function( value, tbl )
			if( 'on' == value ) then
				return 'YES'
			elseif ( nil ~= tbl.poster	and
					(	'1pageinto2sheets' == tbl.poster		or
						'1pageinto4sheets' == tbl.poster		or
						'1pageinto9sheets' == tbl.poster			) )then
				return 'YES'
			else
				return 'NO'
			end
		end,
		default = { 'EDGETOEDGE', 'NO' }
	},

	edgesmoothing = {
		item = 'SMOOTHING',
		values = {
			on	= 'ON',
			off	= 'OFF'
		},
		default = { 'SMOOTHING', 'OFF' }
	},

	fusersetting = {
		item = 'PRINTHEADDIRECTION',
		values = {
			standard	= 'AUTO',
			highquality	= 'UNIDIRECTIONAL'
		},
		default = { 'PRINTHEADDIRECTION', 'AUTO' }
	},

	tonersaving = {
		item = function( value, tbl )
			if( nil == tbl.economycolormode ) then
				return 'ECONOMODE'
			else
				return nil
			end
		end,
		values = function( value, tbl )
			if( 'on' == value ) then
				return 'ON'
			elseif( 'tonersave1' == value ) then
				return 'MODE1'
			elseif( 'tonersave2' == value ) then
				return 'MODE2'
			else
				return 'OFF'
			end
		end
	},

	economycolormode = {
		item = 'ECONOMODE',
		values = {
			on	= 'ON',
			off	= 'OFF'
		},
		default = { 'ECONOMODE', 'OFF' }
	},

	enhancedtonersaving = {
		item = {'ECONOMODEFILL', 'ECONOMODELINE', 'ECONOMODETEXT', 'ECONOMODEIMAGE', 'ECONOMODE'},
		values = {
			light		= {'MODE2'	, 'MODE2'	, 'OFF'	, 'MODE2'	, nil	},
			mediumlight	= {'MODE1'	, 'MODE1'	, 'OFF'	, 'MODE1'	, nil	},
			off			= {'OFF'	, 'OFF'		, 'OFF'	, 'OFF'		, 'OFF'	}
		},
		default = {
			{ 'ECONOMODEFILL'	, 'ECONOMODELINE'	, 'ECONOMODETEXT'	, 'ECONOMODEIMAGE'	, 'ECONOMODE'	},
			{ 'OFF'				, 'OFF'				, 'OFF'				, 'OFF'				, 'OFF'			}
		}
	},

	blackoverprint = {
		item = 'BLACKOVERPRINT',
		values = function(value, tbl)
			if( '1' == tbl.boptype ) then
				if( 'cmyk' == tbl.grayreproduction ) then
					return 'OFF'
				else
					if( 'on' == value ) then
						return 'ON'
					else
						return 'OFF'
					end
				end
			else
				return nil
			end
		end
	},

	staple = {
		item = {'STAPLE'},
		values = STAPLE_TABLE,
		default = { 'STAPLE', 'OFF' }
	},

	multisheetfold = {
		item = 'PILINGFOLD',
		values = function( value, tbl )
			if( 'on' == value ) then
				return 'ON'
			else
				return 'OFF'
			end
		end,
	},

	ringbinding = {
		item = function( value, tbl )
			local t = {
				off					= 'RINGBINDMODE',
				leftortop			= { 'RINGBINDMODE'	, 'BINDPOSITION' },
				rightortop			= { 'RINGBINDMODE'	, 'BINDPOSITION' },
				leftortoppunchonly	= { 'RINGBINDMODE'	, 'BINDPOSITION' },
				rightortoppunchonly	= { 'RINGBINDMODE'	, 'BINDPOSITION' }
			}
			return t[value] and t[value] or 'RINGBINDMODE'
		end,
		values = function(value, tbl)
			local isPortraitValue = isPortrait( tbl )
			local t = {
				off					= 'OFF',
				leftortop			= { 'RING'	, (function() if isPortraitValue then return 'LEFTPORT' else return 'TOPLAND' end end )() },
				rightortop			= { 'RING'	, (function() if isPortraitValue then return 'RIGHTPORT' else return 'TOPLAND' end end )() },
				leftortoppunchonly	= { 'PUNCH'	, (function() if isPortraitValue then return 'LEFTPORT' else return 'TOPLAND' end end )() },
				rightortoppunchonly	= { 'PUNCH'	, (function() if isPortraitValue then return 'RIGHTPORT' else return 'TOPLAND' end end )() }
			}
			return t[value] and t[value] or 'OFF'
		end
	},

	twinloopbinding = {
		item = function( value, tbl )
			local t = {
				off					= 'TWINLOOPBINDMODE',
				leftortop			= { 'TWINLOOPBINDMODE'	, 'TWINLOOPPOSITION' },
				rightortop			= { 'TWINLOOPBINDMODE'	, 'TWINLOOPPOSITION' }
			}
			return t[value] and t[value] or 'TWINLOOPBINDMODE'
		end,
		values = function(value, tbl)
			local isPortraitValue = isPortrait( tbl )
			local t = {
				off					= 'OFF',
				leftortop			= { 'ON'	, (function() if isPortraitValue then return 'LEFTPORT' else return 'TOPLAND' end end )() },
				rightortop			= { 'ON'	, (function() if isPortraitValue then return 'RIGHTPORT' else return 'TOPLAND' end end )() }
			}
			return t[value] and t[value] or 'OFF'
		end
	},

	perfectbinding = {
		item = 'PERFECTBIND',
		values = function(value, tbl)
			if( 'on' == value ) then
				return 'ON'
			else
				return 'OFF'
			end
		end
	},

	perfectbinding_direction = {
		item = 'BINDPOSITION',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				local isPortraitValue = isPortrait( tbl )
				
				if(	nil ~= tonumber(tbl.jh_document_paperwidth_mm) and
					nil ~= tonumber(tbl.jh_document_paperheight_mm) and
					tonumber(tbl.jh_document_paperwidth_mm) > tonumber(tbl.jh_document_paperheight_mm) ) then
					isPortraitValue = not isPortraitValue
				end
				
				local t ={
					leftortop	= (function() if isPortraitValue then return 'LEFTPORT' else return 'TOPLAND' end end )(),
					rightortop	= (function() if isPortraitValue then return 'RIGHTPORT' else return 'TOPLAND' end end )()
				}
				return t[value] and t[value] or t['leftortop']
			end
		end
	},

	usercode = {
		item = 'USERCODE',
		values = function(value, tbl)
			if( '2' == tbl.controllerplatformtype) then
				return nil
			else
				if( isWrongString( tbl.usercode, 231 ) ) then
					return '""'
				else
					return '"' .. tbl.usercode .. '"'
				end
			end
		end
	},

	jobname = {
		item = 'JOBNAME',
		values = function(value, tbl)
			return GetJobName( tbl )
		end
	},

	jh_trackid = {
		item = 'TRACKID',
		values = function(value, tbl)
			return '"' .. get_trimmed_command_string( tbl.jh_trackid, GetCommandCharcode(tbl), 20 ) .. '"'
		end
	},

	windowslogonusername = {
		item = 'HOSTLOGINNAME',
		values = function(value, tbl)
			return '"' .. get_trimmed_command_string( tbl.windowslogonusername, GetCommandCharcode(tbl), 20 ) .. '"'
		end
	},

	billingcode = {
		item = 'BILLINGCODE',
		values = function(value, tbl)
			if( isMember( tbl.jobtype, { 'documentserver', 'documentserverrgb' } ) ) then
				return nil
			else
				if isWrongString( tbl.billingcode, 228 ) then
					return '""'
				else
					return '"' .. tbl.billingcode .. '"'
				end
			end
		end
	},

	standardpages_inputtray = {
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return nil
			else
				if( 'on' == tbl.perfectbinding ) then
					return { 'AUTOTRAYCHANGE', 'TRAY', 'PERFECTBINDPAPERTRAY' }
				else
					return { 'AUTOTRAYCHANGE', 'TRAY', nil }
				end
			end
		end,
		values = {
			autotrayselect		= {'ON','ALL','ALL'},
			tray1				= {'OFF','TRAY1','TRAY1'},
			tray2				= {'OFF','TRAY2','TRAY2'},
			tray3				= {'OFF','TRAY3','TRAY3'},
			tray3lct			= {'OFF','LCT','LCT'},
			tray4				= {'OFF','TRAY4','TRAY4'},
			tray5				= {'OFF','TRAY5','TRAY5'},
			tray6				= {'OFF','TRAY6','TRAY6'},
			tray7				= {'OFF','TRAY7','TRAY7'},
			tray8				= {'OFF','TRAY8','TRAY8'},
			tray9				= {'OFF','TRAY9','TRAY9'},
			tray10				= {'OFF','TRAY10','TRAY10'},
			tray11				= {'OFF','TRAY11','TRAY11'},
			traya				= {'OFF','TRAYA','TRAYA'},
			bypasstray			= {'OFF','BYPASS','BYPASS'},
			onesheetbypass		= {'OFF','TRAY7','TRAY7'},
			frontbypasstray		= {'OFF','BYPASS','BYPASS'},
			largecapacitytray	= {'OFF','LCT','LCT'}
		}
	},

	transparency_inputtray = {
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return { 'AUTOTRAYCHANGE', 'TRAY' }
			else
				return nil
			end
		end,
		values = {
			autotrayselect		= {'ON','ALL'},
			tray1				= {'OFF','TRAY1'},
			tray2				= {'OFF','TRAY2'},
			tray3				= {'OFF','TRAY3'},
			tray3lct			= {'OFF','LCT'},
			tray4				= {'OFF','TRAY4'},
			tray5				= {'OFF','TRAY5'},
			tray6				= {'OFF','TRAY6'},
			tray7				= {'OFF','TRAY7'},
			tray8				= {'OFF','TRAY8'},
			tray9				= {'OFF','TRAY9'},
			tray10				= {'OFF','TRAY10'},
			tray11				= {'OFF','TRAY11'},
			traya				= {'OFF','TRAYA'},
			bypasstray			= {'OFF','BYPASS'},
			onesheetbypass		= {'OFF','TRAY7'},
			frontbypasstray		= {'OFF','BYPASS'},
			largecapacitytray	= {'OFF','LCT'}
		}
	},

	standardpages_tabstockprinting = {
		item = 'TABPOSITION',
		values = {
			off	= 'OFF',
			on	= 'RIGHT'
		}
	},

	standardpages_shiftprintposition = {
		item = function(value,tbl)
			if( 'on' == tbl.standardpages_tabstockprinting ) then
				return 'TABSTOCKSHIFTPRINT'
			else
				return nil
			end
		end,
		values = {
			off	= 'OFF',
			on	= 'ON'
		}
	},
	
	letterheadmode = {
		item = 'LETTERHEADSETTING',
		values = function( value, tbl )
			if( 'off' == tbl.letterheadmode )then
					return 'OFF'
			elseif( 'auto' == tbl.letterheadmode )then
				return 'AUTO'
			elseif( 'always' == tbl.letterheadmode )then
				return 'ALWAYS'
			else
				return nil
			end
		end
	},
	
	duplexmode = {
		item = 'DUPLEXMODE',
		values = function( value, tbl )
			if( 'on' == value ) then
				return 'ON'
			elseif( 'off' == value )then
				if( 'off' == tbl.letterheadmode ) then
					return nil
				else
					return 'OFF'
				end
			else
				return nil
			end
		end
	},

	frontcover =	{
		item = 'FRONTCOVERPRINT',
		values = function(value, tbl)
			if( 'off' == value or isFrontCoverWithBooklet( tbl ) ) then
				return 'OFF'
			else
				return nil
			end
		end
	},

	frontcover_printmode =	{
		item = function(value,tbl)
			if( isFrontCoverWithBooklet( tbl ) ) then
				return 'BACKCOVERPRINT'
			elseif( 'on' == tbl.frontcover ) then
				return  'FRONTCOVERPRINT'
			else
				return nil
			end
		end,
		values = function(value, tbl)
			if( isFrontCoverWithBooklet( tbl ) ) then
				if( '3' == tbl.centerstaple_swapcover ) then
					return {
						leaveblank	= 	'BLANK',
						oneside		= 	'INNERCOPY',
						bothsides	= 	'INNERCOPY'
					}
				elseif( isMember( tbl.centerstaple_swapcover, {'1','2'} ) ) then
					return {
						leaveblank	= 	'BLANK',
						oneside		= 	'OUTERCOPY',
						bothsides	= 	'OUTERCOPY'
					}
				end
			elseif( 'on' == tbl.frontcover ) then
				return {
					leaveblank	= 	'BLANK',
					oneside		= 	'SIMPLEXCOPY',
					bothsides	= 	'DUPLEXCOPY'
				}
			else
				return nil
			end
		end
	},

	frontcover_inputtray =	{
		item = function( value, tbl )
			if( isFrontCoverWithBooklet( tbl ) and '1' == tbl.centerstaple_swapcover ) then
				return 'BACKCOVERPRINTTRAY'
			elseif( 'on' == tbl.frontcover ) then
				return 'FRONTCOVERPRINTTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	backcover =	{
		item = 'BACKCOVERPRINT',
		values = function(value, tbl)
			if( 'off' == value and not isFrontCoverWithBooklet(tbl) ) then
				return 'OFF'
			else
				return nil
			end
		end
	},

	backcover_printmode =	{
		item = function(value,tbl)
			if( 'on' == tbl.backcover and not isFrontCoverWithBooklet(tbl) ) then
				return  'BACKCOVERPRINT'
			else
				return nil
			end
		end,
		values = {
			leaveblank			= 'BLANK',
			printonbackcover 	= 'INNERCOPY',
			printonouter 		= 'OUTERCOPY'
		}
	},

	backcover_inputtray =	{
		item = function( value, tbl )
			if( 'on' == tbl.backcover and not isFrontCoverWithBooklet(tbl) ) then
				return 'BACKCOVERPRINTTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	slipsheet = {
		item = 'SLIPSHEETPRINT',
		values = function( value, tbl )
			if( 'off' == value ) then
				return 'OFF'
			else
				if( 'printonslipsheet' == tbl.slipsheet_printmode ) then
					return 'COPY'
				else
					return 'BLANK'
				end
			end
		end
	},


	slipsheet_inputtray =	{
		item = function( value, tbl )
			if( 'on' == tbl.slipsheet ) then
				return 'SLIPSHEETPRINTTRAY'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	chapterpages = {
		item =	'CHAPTERPRINT',
		values ={
			on	= 'ON'
		}

	},

	jh_chaptertray1 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY1'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode1 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX1'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray2 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY2'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode2 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX2'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray3 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY3'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode3 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX3'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray4 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY4'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode4 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX4'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray5 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY5'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode5 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX5'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray6 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY6'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode6 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX6'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray7 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY7'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode7 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX7'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray8 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY8'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode8 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX8'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray9 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY9'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode9 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX9'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_chaptertray10 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTTRAY10'
			else
				return nil
			end
		end,
		values = INPUTTRAY_TABLE
	},

	jh_chaptermode10 = {
		item =	function(value,tbl)
			if( 'on' == tbl.chapterpages ) then
				return 'CHAPTERPRINTDUPLEX10'
			else
				return nil
			end
		end,
		values = {
			leaveblank 	= 'OFF',
			oneside		= 'OFF',
			bothsides	= 'ON'
		}
	},

	jh_charset = {
		item =  { 'HOSTCHARSET2', 'DISPCHARSET2' },
		values = function(value, tbl)
			if( 'utf8' == GetCommandCharcode(tbl)) then
				return { '33', '33' }
			elseif( isMember( value, { '1', '2', '3', '4', '5', '6', '8', '9', '11', '12', '13' } )) then
				return { value, value }
			else
				return { '2', '2' }
			end
		end
	},

	loginusername = {
		item = function( value, tbl )
			local luname = get_trimmed_command_string( value, GetCommandCharcode(tbl), 217 )
			if( 'on' == tbl.userauthentication and not isEmpty( luname ) ) then
				return { 'AUTHENTICATIONUSERNAME', 'AUTHENTICATIONUSERNAMECHARSET', 'AUTHENTICATIONPASSWORDCHARSET' }
			else
				return { 'AUTHENTICATIONUSERNAME', 'AUTHENTICATIONUSERNAMECHARSET'}
			end
		end,
		values = function(value, tbl)
			local luname = get_trimmed_command_string( value, GetCommandCharcode(tbl), 217 )
			if( 'on' == tbl.userauthentication and not isEmpty( luname ) ) then
				if ( 'utf8' == GetCommandCharcode(tbl) ) then
					return { '"' .. luname .. '"', '33', '33'}
				elseif( isMember( tbl.jh_charset, { '1', '2', '3', '4', '5', '6', '8', '9', '11', '12', '13' } )) then
					return { '"' .. luname .. '"', tbl.jh_charset, tbl.jh_charset }
				else
					return { '"' .. luname .. '"', '2', '2' }
				end
			else
				if ( 'utf8' == GetCommandCharcode(tbl) ) then
					return { '""', '33' }
				elseif( isMember( tbl.jh_charset, { '1', '2', '3', '4', '5', '6', '8', '9', '11', '12', '13' } )) then
					return { '""', tbl.jh_charset }
				else
					return { '""', '2' }
				end
			end
		end
	},

	jh_loginpassword3 = {
		item = { 'AUTHENTICATIONPASSWORD', 'AUTHENTICATIONENCRYPTSETTING', 'AUTHENTICATIONENCRYPTMETHOD' },
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication ) then
				local luname = get_trimmed_command_string( tbl.loginusername, GetCommandCharcode(tbl), 217 )
				if( isEmpty( luname ) ) then
					return nil
				else
					if( '5' == tbl.jobpasswordgeneration ) then
						return { '"' .. value .. '"', '2', '10' }
					else
						if( 'on' == tbl.withencryption ) then
							return { '"' .. value .. '"', '2', '0' }
						else
							return { '"' .. value .. '"', '1', '0' }
						end
					end
				end
			else
				return nil
			end
		end
	},

	jh_loginpassword4 = {
		item = { 'AUTHENTICATIONPASSWORD', 'AUTHENTICATIONENCRYPTSETTING', 'AUTHENTICATIONENCRYPTMETHOD' },
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication ) then
				local luname = get_trimmed_command_string( tbl.loginusername, GetCommandCharcode(tbl), 217 )
				if( isEmpty( luname ) ) then
					return nil
				else
					if( '5' == tbl.jobpasswordgeneration ) then
						return { '"' .. value .. '"', '2', '10' }
					else
						if( 'on' == tbl.withencryption ) then
							return { '"' .. value .. '"', '2', '0' }
						else
							return { '"' .. value .. '"', '1', '0' }
						end
					end
				end
			else
				return nil
			end
		end
	},

	jh_loginpassword5 = {
		item = { 'AUTHENTICATIONPASSWORD', 'AUTHENTICATIONENCRYPTSETTING', 'AUTHENTICATIONENCRYPTMETHOD' },
		values = function(value, tbl)
			if( 'on' == tbl.userauthentication ) then
				local luname = get_trimmed_command_string( tbl.loginusername, GetCommandCharcode(tbl), 217 )
				if( isEmpty( luname ) ) then
					return nil
				else
					if( '5' == tbl.jobpasswordgeneration ) then
						return { '"' .. value .. '"', '2', '10' }
					else
						if( 'on' == tbl.withencryption ) then
							return { '"' .. value .. '"', '2', '0' }
						else
							return { '"' .. value .. '"', '1', '0' }
						end
					end
				end
			else
				return nil
			end
		end
	},

	joboffset = {
		item = 'JOBOFFSET',
		values = function(value, tbl)
			if( isCollate(tbl) ) then
				return {
						off		= 'OFF',
						rotate	= 'ROTATE',
						shift	= 'SHIFT'
				}
			else
				return 'OFF'
			end
		end,
		default = { 'JOBOFFSET', 'OFF' }
	},

	outputtray = {
		item = 'OUTBIN',
		values = OUTPUTTRAY_TABLE
	},

	perfectbinding_imgpos_along_mm = {
		item		= { 'IMAGEVERTADJUSTMENT' },
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	perfectbinding_imgpos_across_mm = {
		item		= { 'IMAGEHORIZADJUSTMENT' },
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

-- perfectbinding_customsize_width_mm
	 jh_print_paperwidth_mm = {
		item		= { 'PERFECTBINDPAPERCUSTOMWIDTH' },
		values	= function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'CUSTOM' == tbl.jh_perfectbinding_paper_jobcommand and
				'custompapersize' ~= getPrintSize( tbl ) ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	custompapersize_width_mm = {
		item		= { 'PERFECTBINDPAPERCUSTOMWIDTH' },
		values	= function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'custompapersize' == getPrintSize( tbl ) ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

-- perfectbinding_customsize_length_mm
	jh_print_paperheight_mm = {
		item		= { 'PERFECTBINDPAPERCUSTOMLENGTH' },
		values	= function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'CUSTOM' == tbl.jh_perfectbinding_paper_jobcommand and
				'custompapersize' ~= getPrintSize( tbl ) ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

	custompapersize_length_mm = {
		item		= { 'PERFECTBINDPAPERCUSTOMLENGTH' },
		values	= function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'custompapersize' == getPrintSize( tbl ) ) then
				return GetMMValue( value )
			else
				return nil
			end
		end
	},

--perfectbinding_printsize
	jh_perfectbinding_paper_jobcommand = {
		item = 'PERFECTBINDPAPERSIZE',
		values =  function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return value
			else
				return nil
			end
		end
	},

	punch = {
		item = {
			['off']					= { 'PUNCH' },
			['left']				= { 'PUNCH'	},
			['top']					= { 'PUNCH'	},
			['right']				= { 'PUNCH'	},
			['1atlefttop']			= { 'PUNCH' },
			['2atleft']				= { 'PUNCH' },
			['2attop']				= { 'PUNCH' },
			['2atright']			= { 'PUNCH' },
			['3atleft']				= { 'PUNCH' },
			['3attop']				= { 'PUNCH' },
			['3atright']			= { 'PUNCH' },
			['4atleft']				= { 'PUNCH' },
			['4attop']				= { 'PUNCH' },
			['4atright']			= { 'PUNCH' },
			['multiatleftortop']	= { 'PUNCH', 'PUNCHHOLE' },
			['multiatrightortop']	= { 'PUNCH', 'PUNCHHOLE' },
			['multiatleft']			= { 'PUNCH', 'PUNCHHOLE' },
			['multiattop']			= { 'PUNCH', 'PUNCHHOLE' },
			['multiatright']		= { 'PUNCH', 'PUNCHHOLE' },
			['doubleatleftortop']	= { 'PUNCH', 'PUNCHHOLE' },
			['doubleatrightortop']	= { 'PUNCH', 'PUNCHHOLE' },
		},
		values = function(value, tbl)
			local isPortraitValue = isPortrait( tbl )
			return {
				['off']					= { 'OFF'		},
				['left']				= { 'LEFTPORT'	},
				['top']					= { 'TOPPORT'	},
				['right']				= { 'RIGHTPORT'	},
				['1atlefttop']			= { 'LEFTTOP'	},
				['2atleft']				= { 'LEFTPORT'	},
				['2attop']				= { 'TOPPORT'	},
				['2atright']			= { 'RIGHTPORT'	},
				['3atleft']				= { 'LEFTPORT'	},
				['3attop']				= { 'TOPPORT'	},
				['3atright']			= { 'RIGHTPORT'	},
				['4atleft']				= { 'LEFTPORT'	},
				['4attop']				= { 'TOPPORT'	},
				['4atright']			= { 'RIGHTPORT'	},
				['multiatleftortop']	= { (function() if isPortraitValue then return 'LEFTPORT' else return 'TOPPORT' end end)(),		'MULTI'		},
				['multiatrightortop']	= { (function() if isPortraitValue then return 'RIGHTPORT' else return 'TOPPORT' end end)(),	'MULTI'		},
				['multiatleft']			= { 'LEFTPORT',																					'MULTI'		},
				['multiattop']			= { 'TOPPORT',																					'MULTI'		},
				['multiatright']		= { 'RIGHTPORT',																				'MULTI'		},
				['doubleatleftortop']	= { (function() if isPortraitValue then return 'TOPPORT' else return 'LEFTPORT' end end)(),		'DOUBLE'	},
				['doubleatrightortop']	= { (function() if isPortraitValue then return 'TOPPORT' else return 'RIGHTPORT' end end)(),	'DOUBLE'	}
			}
		end,
		default = { 'PUNCH', 'OFF' }
	},

	punchunit = {
		item = 'PUNCHHOLE',
		values = function(value, tbl)
			local punch_value = tbl.punch

			return {
				['none']				= nil,
				['1holetypejp']			=	({
												['1atlefttop']			= 'JP1'
												})[punch_value],
				['2holestypejp']		=	({
												['2atleft']				= 'JP2',
												['2attop']				= 'JP2',
												['2atright']			= 'JP2',
												})[punch_value],
				['2holestypeeu']		=	({
												['2atleft']				= 'JP2',
												['2attop']				= 'JP2',
												['2atright']			= 'JP2',
												})[punch_value],
				['2holestypeus2']		=	({
												['2atleft']				= 'US2',
												['2attop']				= 'US2',
												['2atright']			= 'US2',
												})[punch_value],
				['2and3holestypejp']	=	({
												['2atleft']				= 'JP2',
												['2attop']				= 'JP2',
												['2atright']			= 'JP2',
												['3atleft']				= 'US3',
												['3attop']				= 'US3',
												['3atright']			= 'US3',
												})[punch_value],
				['2and3holestypeus']	=	({
												['2atleft']				= 'US2',
												['2attop']				= 'US2',
												['2atright']			= 'US2',
												['3atleft']				= 'US3',
												['3attop']				= 'US3',
												['3atright']			= 'US3',
												})[punch_value],
				['2and4holestypeeu']	=	({
												['2atleft']				= 'JP2',
												['2attop']				= 'JP2',
												['2atright']			= 'JP2',
												['4atleft']				= 'EU4',
												['4attop']				= 'EU4',
												['4atright']			= 'EU4',
												})[punch_value],
				['3holestypeus3']		=	({
												['3atleft']				= 'US3',
												['3attop']				= 'US3',
												['3atright']			= 'US3',
												})[punch_value],
				['4holestypene']		=	({
												['4atleft']				= 'NEU4',
												['4attop']				= 'NEU4',
												['4atright']			= 'NEU4',
												})[punch_value],
				['4holestypeeu']		=	({
												['4atleft']				= 'EU4',
												['4attop']				= 'EU4',
												['4atright']			= 'EU4',
												})[punch_value]
			}
		end
	},

	punchunit_locale = {
		item = 'PUNCHHOLE',
		values = function(value, tbl)
			local punch_value = tbl.punch

			return {
				['europe']	=	({
												['2atleft']				= 'JP2',
												['2attop']				= 'JP2',
												['2atright']			= 'JP2',
												['4atleft']				= 'EU4',
												['4attop']				= 'EU4',
												['4atright']			= 'EU4',
												})[punch_value],
				['northamerica']	=	({
												['2atleft']				= 'US2',
												['2attop']				= 'US2',
												['2atright']			= 'US2',
												['3atleft']				= 'US3',
												['3attop']				= 'US3',
												['3atright']			= 'US3',
												})[punch_value],
				['northeurope']	=	({
												['4atleft']				= 'NEU4',
												['4attop']				= 'NEU4',
												['4atright']			= 'NEU4',
												})[punch_value],
				['japan']		=	({
												['2atleft']				= 'JP2',
												['2attop']				= 'JP2',
												['2atright']			= 'JP2',
												['3atleft']				= 'US3',
												['3attop']				= 'US3',
												['3atright']			= 'US3',
												})[punch_value],
			}
		end
	},

	perfectbinding_cuttype = {
		item		= 'TRIMMODE',
		values	= function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				local t = {
					donotcut						= 'OFF',
					cutforeedgespecifyfinishingsize	= 'SINGLECUTFINISHSIZE',
					cutforeedge						= 'SINGLECUT',
					cut3edgesspecifyfinishingsize	= 'TRIPLECUTFINISHSIZE',
					cut3edges						= 'TRIPLECUT'
				}

				return t[value] and t[value] or t['donotcut']
			else
				return nil
			end
		end
	},

--perfectbinding_cutpos_finishing_size
	jh_perfectbinding_cutpos_paper_jobcommand = {
		item = { 'TRIPLECUTFINISHSIZE' },
		values =  function(value, tbl)
			if( 'on' == tbl.perfectbinding and 'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype) then
				return value
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_head_mm = {
		item = 'TOPCUT',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and  'cut3edges' == tbl.perfectbinding_cuttype) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_tail_mm = {
		item = 'BOTTOMCUT',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding  and 'cut3edges' == tbl.perfectbinding_cuttype) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_fore_mm = {
		item = 'SIDEEDGECUT',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and ( 'cutforeedge' == tbl.perfectbinding_cuttype or 'cut3edges' == tbl.perfectbinding_cuttype ) ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_customsize_width_mm = {
		item = 'TRIPLECUTFINISHSIZECUSTOMWIDTH',
		values = function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and
				'custompapersize' == tbl.perfectbinding_cutpos_finishing_size ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	jh_perfectbinding_cutpos_paperwidth_mm = {
		item = 'TRIPLECUTFINISHSIZECUSTOMWIDTH',
		values = function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and
				'CUSTOM' == tbl.jh_perfectbinding_cutpos_paper_jobcommand and
				'custompapersize' ~= tbl.perfectbinding_cutpos_finishing_size ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_customsize_length_mm = {
		item = 'TRIPLECUTFINISHSIZECUSTOMLENGTH',
		values = function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and
				'custompapersize' == tbl.perfectbinding_cutpos_finishing_size ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	jh_perfectbinding_cutpos_paperheight_mm = {
		item = 'TRIPLECUTFINISHSIZECUSTOMLENGTH',
		values = function(value, tbl)
			if(	'on' == tbl.perfectbinding and
				'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype and
				'CUSTOM' == tbl.jh_perfectbinding_cutpos_paper_jobcommand and
				'custompapersize' ~= tbl.perfectbinding_cutpos_finishing_size ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_length_mm = {
		item = 'SINGLECUTFINISHSIZELENGTH',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and 'cutforeedgespecifyfinishingsize' == tbl.perfectbinding_cuttype ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	perfectbinding_inputtray =	{
		item = 'PERFECTBINDCOVERTRAY',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding ) then
				return INPUTTRAY_TABLE
			else
				return nil
			end
		end
	},

	perfectbinding_cutpos_along_mm = {
		item = 'FINISHSIZEVERTADJUSTMENT',
		values = function(value, tbl)
			if( 'on' == tbl.perfectbinding and 'cut3edgesspecifyfinishingsize' == tbl.perfectbinding_cuttype ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	trimming = {
		item	= 'TRIMTRIMMODE',
		values	= {
			on	= 'SINGLECUT',
			off	= 'OFF'
		},
		default	= { 'TRIMTRIMMODE', 'OFF' }
	},

	trimming_cutedge_mm = {
		item = 'TRIMSIDEEDGECUT',
		values = function(value, tbl)
			if( 'on' == tbl.trimming and
				'1' ~= tbl.disable_cutedge_option ) then
				return GetMMValue(value)
			else
				return nil
			end
		end
	},

	genericfunction = {
		item	= 'GENERICFUNC',
		values	= {
			off					= 'OFF',
			genericfunction1	= 'FUNC1',
			genericfunction2	= 'FUNC2',
			genericfunction3	= 'FUNC3',
			genericfunction4	= 'FUNC4',
			genericfunction5	= 'FUNC5',
			genericfunction6	= 'FUNC6',
			genericfunction7	= 'FUNC7',
			genericfunction8	= 'FUNC8',
			genericfunction9	= 'FUNC9'
		}
	},

	squarefold = {
		item	= 'SQUAREFOLD',
		values	= {
			off	= 'OFF',
			on	= 'ON'
		}
	},

	hostmachinename = {
		item = 'HOSTNAME',
		values = function( value, tbl )
			return '"' ..  get_trimmed_command_string( value, GetCommandCharcode(tbl), 63 )  .. '"'
		end
	},

	portname = {
		item = 'HOSTPORTNAME',
		values = function( value, tbl )
			return '"' ..  get_trimmed_command_string( value, GetCommandCharcode(tbl), 64 )  .. '"'
		end
	},

	printericonname = {
		item = 'HOSTPRINTERNAME',
		values = function( value, tbl )
			return '"' ..  get_trimmed_command_string( value, GetCommandCharcode(tbl), 64 )  .. '"'
		end
	},

	cc_print = {
		item	= 'PRINTERMARKINGTYPE',
		values	= function( value )
			if( 'on' ~= value ) then
				return 'OFF'
			end
		end
	},

	cc_type = {
		item	= 'PRINTERMARKINGTYPE',
		values	= function( value, tbl )
			if( 'on' == tbl.cc_print ) then
				if( 'copyguard' == value ) then
					return 'COPYGUARD'
				else
					return 'COPYINHIBITION'
				end
			end
		end
	},

	cc_preventioncopypattern_color = {
		item	= 'UCPPATTERNCOLOR',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local t = {
					black	= 'BLACK',
					cyan	= 'CYAN',
					magenta	= 'MAGENTA'
				}
				return t[ value ] and t[ value ] or t.black
			end
		end
	},

	cc_preventioncopypattern_density = {
		item	= 'UCPPATTERNDENSITY',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local t = {
					verylight	= '1',
					light		= '2',
					middle		= '3',
					dark		= '4',
					verydark	= '5'
				}
				return t[ value ] and t[ value ] or t.middle
			end
		end
	},

	cc_preventioncopypattern_masktype = {
		item	= 'UCPMASKTYPE',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local t = {
					none				= 'NONE',
					wavecrest			= 'SEIGAIHA',
					mesh				= 'AMIME',
					lattice1			= 'KOUSHI1',
					lattice2			= 'KOUSHI2',
					interlockingcircles	= 'SHIPPOU',
					shokkoh				= 'SYOKKOU',
					matsukawabishi		= 'MATSUKAWABISHI',
					scale				= 'UROKO',
					higaki				= 'HIGAKI',
					hexagonal			= 'KIKKOU'
				}
				return t[ value ] and t[ value ] or t.none
			end
		end
	},

	cc_preventioncopypattern_patterneffects = {
		item	= 'UCPEFFECTCOPYINHIBITION',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local t = {
					normal								= 'ALL',
					reversepatterns						= 'REVERSE',
					reversepatternsandbackgroundonly	= 'BACKGROUND',
					textonly							= 'STRING'
				}
				return t[ value ] and t[ value ] or t.normal
			end
		end
	},

	cc_preventioncopypattern_text = {
		item	= 'UCPSTRING',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local userdefined	= not isEmpty( value ) and value or 'COPY'
				local jobname		= not isEmpty( tbl.jobname ) and tbl.jobname or 'jobname'
				local username 		= not isEmpty( tbl.windowslogonusername ) and tbl.windowslogonusername or 'username'
				local jobtime		= generate_date_and_time()
				if( '2000/01/01 00:00' == jobtime ) then
					jobtime = 'YYYY/MM/DD hh:mm'
				end

				local t = {
					userdefined									= userdefined,
					loginname									= username,
					printjobname								= jobname,
					printjobtime								= jobtime,
					loginnameandprintjobname					= username .. ' ' .. jobname,
					loginnameandprintjobtime					= username .. ' ' .. jobtime,
					printjobnameandprintjobtime					= jobname .. ' ' .. jobtime,
					loginnameandprintjobnameandprintjobtime		= username .. ' ' .. jobname .. ' ' .. jobtime,
				}

				local v = tbl.cc_preventioncopypattern_texttype and tbl.cc_preventioncopypattern_texttype or 'userdefined'
				return '"' .. get_trimmed_command_string( t[ v ] and t[ v ] or t.userdefined , 'utf8', 64 ) .. '"'
			end
		end
	},

	cc_preventioncopypattern_font = {
		item	= 'LPARM:PCL UCPFONTNUMBER',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local t = {
					['courier']							= '0',
					['cg-times']						= '1',
					['cg-times-bold']					= '2',
					['cg-times-italic']					= '3',
					['cg-times-bold-italic']			= '4',
					['cg-omega']						= '5',
					['cg-omega-bold']					= '6',
					['cg-omega-italic']					= '7',
					['cg-omega-bold-italic']			= '8',
					['coronet']							= '9',
					['clarendon-condensed']				= '10',
					['univers-medium']					= '11',
					['univers-bold']					= '12',
					['univers-medium-italic']			= '13',
					['univers-bold-italic']				= '14',
					['univers-medium-condensed']		= '15',
					['univers-bold-condensed']			= '16',
					['univers-medium-condensed-italic']	= '17',
					['univers-bold-condensed-italic']	= '18',
					['antiqueolive']					= '19',
					['antiqueolive-bold']				= '20',
					['antiqueolive-italic']				= '21',
					['garamond-antiqua']				= '22',
					['garamond-halbfett']				= '23',
					['garamond-kursiv']					= '24',
					['garamond-kursiv-halbfett']		= '25',
					['marigold']						= '26',
					['albertus-medium']					= '27',
					['albertus-extra-bold']				= '28',
					['arial']							= '29',
					['arial-bold']						= '30',
					['arial-italic']					= '31',
					['arial-bold-italic']				= '32',
					['times-new-roman']					= '33',
					['times-new-roman-bold']			= '34',
					['times-new-roman-italic']			= '35',
					['times-new-roman-bold-italic']		= '36',
					['symbol']							= '37',
					['wingdings']						= '38',
					['courier-bold']					= '39',
					['courier-italic']					= '40',
					['courier-bold-italic']				= '41',
					['lettergothic']					= '42',
					['lettergothic-bold']				= '43',
					['lettergothic-italic']				= '44',
					['lineprinter8u']					= '45',
					['lineprinter10u']					= '46',
					['lineprinter11u']					= '47',
					['nimbussans-bold']					= '47',
					['lineprinter12u']					= '48',
					['lineprinter0n']					= '49',
					['lineprinter1u']					= '50',
					['nimbusroman-bold']				= '71',
					['nimbusmonol-bold']				= '79'
				}
				return t[ value ] and t[ value ] or t.arial
			end
		end
	},

	cc_preventioncopypattern_position = {
		item	= 'UCPSTRPOSITION',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local t = {
					center			= 'CENTER',
					topleft			= 'TOPLEFT',
					centertop		= 'TOPCENTER',
					topright		= 'TOPRIGHT',
					bottomleft		= 'BOTTOMLEFT',
					centerbottom	= 'BOTTOMCENTER',
					bottomright		= 'BOTTOMRIGHT'
				}
				return t[ value ] and t[ value ] or t.center
			end
		end
	},

	cc_preventioncopypattern_size = {
		item	= 'UCPFONTSIZE',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then
				local val = tonumber( value ) and tonumber( value ) or 70
				val = math.floor( val )

				if( 0 > val or 32767 < val ) then
					val = 70
				end

				return tostring( val )
			end
		end
	},

	cc_preventioncopypattern_angle = {
		item	= 'UCPSTRANGLE',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then

				local t = tonumber( value ) and tonumber( value ) or 30

				if( t < 0 or t > 359 ) then
					t = 30
				end

				return tostring( t )
			end
		end
	},

	cc_preventioncopypattern_repeat = {
		item	= { 'UCPSTRREPEATPRINT', 'UCPLINESPACE' },
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	~= tbl.cc_type	) then

				local val = tonumber( tbl.cc_preventioncopypattern_linespace ) and tonumber( tbl.cc_preventioncopypattern_linespace ) or 70
				val = math.floor( val )

				if( 0 > val or 32767 < val ) then
					val = 70
				end

				local t = {
					['off']								= { 'OFF' },
					['repeat']							= { 'ON', tostring( val ) },
					['repeatandrotatecarriagereturn']	= { 'REVERSELF', tostring( val ) }
				}

				return t[ value ] and t[ value ] or t.off
			end
		end
	},

	cc_copyguard_patterneffects = {
		item	= 'UCPEFFECTCOPYGUARD',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then
				local t = {
					normal			= 'ALL',
					backgroundonly	= 'BACKGROUND'
				}
				return t[ value ] and t[ value ] or t.normal
			end
		end
	},

	cc_copyguard_text = {
		item	= 'UCPSTRING',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then
				local userdefined	= not isEmpty( value ) and value or ''
				local jobname		= not isEmpty( tbl.jobname ) and tbl.jobname or 'jobname'
				local username 		= not isEmpty( tbl.windowslogonusername ) and tbl.windowslogonusername or 'username'
				local jobtime		= generate_date_and_time()
				if( '2000/01/01 00:00' == jobtime ) then
					jobtime = 'YYYY/MM/DD hh:mm'
				end

				local t = {
					userdefined									= userdefined,
					loginname									= username,
					printjobname								= jobname,
					printjobtime								= jobtime,
					loginnameandprintjobname					= username .. ' ' .. jobname,
					loginnameandprintjobtime					= username .. ' ' .. jobtime,
					printjobnameandprintjobtime					= jobname .. ' ' .. jobtime,
					loginnameandprintjobnameandprintjobtime		= username .. ' ' .. jobname .. ' ' .. jobtime,
				}

				local v = tbl.cc_copyguard_texttype and tbl.cc_copyguard_texttype or 'userdefined'
				return '"' .. get_trimmed_command_string( t[ v ] and t[ v ] or t.userdefined, 'utf8', 64 ) .. '"'
			end
		end
	},

	cc_copyguard_font = {
		item	= 'LPARM:PCL UCPFONTNUMBER',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then
				local t = {
					['courier']							= '0',
					['cg-times']						= '1',
					['cg-times-bold']					= '2',
					['cg-times-italic']					= '3',
					['cg-times-bold-italic']			= '4',
					['cg-omega']						= '5',
					['cg-omega-bold']					= '6',
					['cg-omega-italic']					= '7',
					['cg-omega-bold-italic']			= '8',
					['coronet']							= '9',
					['clarendon-condensed']				= '10',
					['univers-medium']					= '11',
					['univers-bold']					= '12',
					['univers-medium-italic']			= '13',
					['univers-bold-italic']				= '14',
					['univers-medium-condensed']		= '15',
					['univers-bold-condensed']			= '16',
					['univers-medium-condensed-italic']	= '17',
					['univers-bold-condensed-italic']	= '18',
					['antiqueolive']					= '19',
					['antiqueolive-bold']				= '20',
					['antiqueolive-italic']				= '21',
					['garamond-antiqua']				= '22',
					['garamond-halbfett']				= '23',
					['garamond-kursiv']					= '24',
					['garamond-kursiv-halbfett']		= '25',
					['marigold']						= '26',
					['albertus-medium']					= '27',
					['albertus-extra-bold']				= '28',
					['arial']							= '29',
					['arial-bold']						= '30',
					['arial-italic']					= '31',
					['arial-bold-italic']				= '32',
					['times-new-roman']					= '33',
					['times-new-roman-bold']			= '34',
					['times-new-roman-italic']			= '35',
					['times-new-roman-bold-italic']		= '36',
					['symbol']							= '37',
					['wingdings']						= '38',
					['courier-bold']					= '39',
					['courier-italic']					= '40',
					['courier-bold-italic']				= '41',
					['lettergothic']					= '42',
					['lettergothic-bold']				= '43',
					['lettergothic-italic']				= '44',
					['lineprinter8u']					= '45',
					['lineprinter10u']					= '46',
					['lineprinter11u']					= '47',
					['nimbussans-bold']					= '47',
					['lineprinter12u']					= '48',
					['lineprinter0n']					= '49',
					['lineprinter1u']					= '50',
					['nimbusroman-bold']				= '71',
					['nimbusmonol-bold']				= '79'
				}
				return t[ value ] and t[ value ] or t.arial
			end
		end
	},

	cc_copyguard_position = {
		item	= 'UCPSTRPOSITION',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then
				local t = {
					center			= 'CENTER',
					topleft			= 'TOPLEFT',
					centertop		= 'TOPCENTER',
					topright		= 'TOPRIGHT',
					bottomleft		= 'BOTTOMLEFT',
					centerbottom	= 'BOTTOMCENTER',
					bottomright		= 'BOTTOMRIGHT'
				}
				return t[ value ] and t[ value ] or t.center
			end
		end
	},

	cc_copyguard_size = {
		item	= 'UCPFONTSIZE',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then
				local val = tonumber( value ) and tonumber( value ) or 70
				val = math.floor( val )

				if( 0 > val or 32767 < val ) then
					val = 70
				end

				return tostring( val )
			end
		end
	},

	cc_copyguard_angle = {
		item	= 'UCPSTRANGLE',
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then

				local t = tonumber( value ) and tonumber( value ) or 30

				if( t < 0 or t > 359 ) then
					t = 30
				end

				return tostring( t )
			end
		end
	},

	cc_copyguard_repeat = {
		item	= { 'UCPSTRREPEATPRINT', 'UCPLINESPACE' },
		values	= function( value, tbl )
			if(	'on'		== tbl.cc_print	and
				'copyguard'	== tbl.cc_type	) then

				local val = tonumber( tbl.cc_copyguard_linespace ) and tonumber( tbl.cc_copyguard_linespace ) or 70
				val = math.floor( val )

				if( 0 > val or 32767 < val ) then
					val = 70
				end

				local t = {
					['off']								= { 'OFF' },
					['repeat']							= { 'ON', tostring( val ) },
					['repeatandrotatecarriagereturn']	= { 'REVERSELF', tostring( val ) }
				}

				return t[ value ] and t[ value ] or t.off
			end
		end
	},

	secretnumbering = {
		item = 'SECRETNUMBERING',
		values = {
			on	= 'ON',
			off	= 'OFF'
		}
	},

	secretnumbering_size = {
		item =	function( value, tbl )
			if( 'on' == tbl.secretnumbering ) then
				return 'SECRETNUMBERINGPRINTSIZE'
			else
				return nil
			end
		end,
		values = {
			small 	= 'SMALL',
			middle	= 'MIDDLE',
			large	= 'LARGE'
		}
	},

	secretnumbering_density = {
		item =	function( value, tbl )
			if( 'on' == tbl.secretnumbering ) then
				return 'SECRETNUMBERINGPRINTDENSITY'
			else
				return nil
			end
		end,
		values = {
			light 		= 'LIGHT',
			middle		= 'MIDDLE',
			littledark	= 'LITTLEDARK',
			dark		= 'DARK'
		}
	},

	secretnumbering_color = {
		item =	function( value, tbl )
			if( 'on' == tbl.secretnumbering ) then
				return 'SECRETNUMBERINGPRINTCOLOR'
			else
				return nil
			end
		end,
		values = {
			black	= 'BLACK',
			red		= 'RED',
			green	= 'GREEN',
			blue	= 'BLUE',
			cyan	= 'CYAN',
			magenta	= 'MAGENTA',
			yellow 	= 'YELLOW'
		}
	},

	secretnumbering_number = {
		item =	function( value, tbl )
			if( 'on' == tbl.secretnumbering ) then
				return 'SECRETNUMBERINGPRINTSTARTNUM'
			else
				return nil
			end
		end,
		values = function( value, tbl )
			return value
		end
	},

	bannerpageprint = {
		item = 'BANNERPAGEPRINT',
		values = {
			on	= 'BEFOREJOB',
			off	= 'OFF'
		}
	},

	jh_bannerpage_mediatype_jobcommand =	{
		item = function( value, tbl )
			if( 'on' == tbl.bannerpageprint ) then
				return 'BANNERPAGEPRINTMEDIA'
			else
				return nil
			end
		end,
		values = function(value, tbl)
			return value
		end
	},

	bannerpageprint_inputtray =	{
		item = function( value, tbl )
			if( 'on' == tbl.bannerpageprint ) then
				return 'BANNERPAGEPRINTTRAY'
			else
				return nil
			end
		end,
		values = function( value, tbl )
			return INPUTTRAY_TABLE[value] and INPUTTRAY_TABLE[value] or INPUTTRAY_TABLE.autotrayselect
		end
	},

	physicalpagecount = {
		item = 'PRINTFACESNUM',
		values = function( value, tbl )
			if( tbl.booklet and 'booklet' == tbl.booklet ) then
				return value
			else
				return nil
			end
		end
	},
}

local function GetCommandCharsetValue( tbl )
	if( 'utf8' == GetCommandCharcode(tbl)) then
		return '33'
	elseif( isMember( value, { '1', '2', '3', '4', '5', '6', '8', '9', '13' } )) then
		return value
	else
		return '2'
	end
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate COMMENT
-----------------------------------------------------------------------------
function generate_pcl_comment(settings_table)

	local fields = {
		'RPJL',
		'WIN2000',
		(function()
			if( 'color' == settings_table.colorspec ) then
				return 'PCL6C'
			else
				return 'PCL6'
			end
		end)(),
		(function()
			if( nil	~= settings_table.driverproductversion and
				''	~= settings_table.driverproductversion		) then
				return settings_table.driverproductversion
			else
				return '1.0.0.0'
			end
		end)(),
		(function()
			if( 'color' == settings_table.colorbw ) then
				return 'COLOR'
			else
				return 'GRAYSCALE'
			end
		end)()
	}

	return '@PJL COMMENT '..fields[1]..','..fields[2]..','..fields[3]..','..fields[4]..','..fields[5]..'\r\n'
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate COMMENT OTHER
-----------------------------------------------------------------------------
function generate_pcl_comment_other(settings_table)
	if( nil ~= settings_table.pclother ) then
		return '@PJL COMMENT OTHER = "' ..settings_table.pclother..'"\r\n'
	else
		return nil
	end
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate BEGIN JOB NAME
-----------------------------------------------------------------------------
function generate_begin_jobname(settings_table)
	return '@PJL JOB NAME = ' .. GetJob_Name( settings_table ) .. '\r\n'
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate JOB ADDITIONAL INFO 
-----------------------------------------------------------------------------
function generate_job_additional_info_userdomain_name(settings_table)
	if( nil ~= settings_table.userdomainname ) then
		return '@PJL SET JOBADDITIONALINFO = ' .. '"UserDomain=' .. get_trimmed_command_string( settings_table.userdomainname, GetCommandCharcode(settings_table), 211 ) .. '"\r\n'
	else
		return nil
	end
end

function generate_job_additional_info_process_name(settings_table)
	if( nil ~= settings_table.jh_process_name ) then
		return '@PJL SET JOBADDITIONALINFO = ' .. '"ProcessName=' .. get_trimmed_command_string( settings_table.jh_process_name, GetCommandCharcode(settings_table), 210 ) .. '"\r\n'
	else
		return nil
	end
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate SCALING in PRINTINFO
-----------------------------------------------------------------------------
function generate_scaling_printinfo(settings_table)
	local scaling_value
	if(	'booklet' == settings_table.booklet	or
		isMember( settings_table.poster, { '1pageinto2sheets', '1pageinto4sheets', '1pageinto9sheets' } )	or
		isMember( settings_table.layout, { '2pagespersheet', '4pagespersheet', '6pagespersheet', '9pagespersheet', '16pagespersheet' } ) )	then
		scaling_value = '3'
	elseif( 'zoom' == settings_table.reduceenlarge and
		nil ~= settings_table.zoomvalue	 ) then

		if( 50 > tonumber(settings_table.zoomvalue)			) then
			scaling_value = '1'
		elseif( 50 	<= 	tonumber(settings_table.zoomvalue) 	and
				100 > 	tonumber(settings_table.zoomvalue) 		) then
			scaling_value = '2'
		elseif(	100 == 	tonumber(settings_table.zoomvalue)		) then
			scaling_value = '3'
		elseif( 100 < 	tonumber(settings_table.zoomvalue) 	and
				200 >= 	tonumber(settings_table.zoomvalue) 		) then
			scaling_value = '4'
		elseif(	200 <	tonumber(settings_table.zoomvalue)		) then
			scaling_value = '5'
		else
			scaling_value = '3'
		end

	elseif( 'fittoprintsize' == settings_table.reduceenlarge 	and
			nil ~= settings_table.jh_print_paperwidth_mm		and
			nil ~= settings_table.jh_print_paperheight_mm		and
			nil ~= settings_table.jh_document_paperwidth_mm		and
			nil ~= settings_table.jh_document_paperheight_mm		) then
		local scaling = 1.0

		local x_scaling = settings_table.jh_print_paperwidth_mm / settings_table.jh_document_paperwidth_mm
		local y_scaling = settings_table.jh_print_paperheight_mm  / settings_table.jh_document_paperheight_mm

		if( x_scaling < y_scaling ) then
			scaling = x_scaling
		elseif( y_scaling < x_scaling)  then
			scaling = y_scaling
		elseif( x_scaling == y_scaling ) then
			scaling = x_scaling
		else
			scaling = 0.1
		end

		if( 	0.5 > 	scaling 		) then
			scaling_value = '1'
		elseif( 0.5 <= 	scaling	and
				1.0 >	scaling 		) then
			scaling_value = '2'
		elseif( scaling == 1.0 			) then
			scaling_value = '3'
		elseif( 1.0 < 	scaling 	and
				2.0 >= 	scaling 		) then
			scaling_value = '4'
		elseif( 2.0 <	scaling			) then
			scaling_value = '5'
		else
			scaling_value = '3'
		end
	else
		scaling_value = '3'
	end
	return 'SCALING='..scaling_value
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate DULPEX_NUP in PRINTINFO
-----------------------------------------------------------------------------
function generate_duplex_nup_printinfo(settings_table)
	local duplex_nup_value

	-- 1 mojime
	duplex_nup_value = '0'
	-- 2 mojime
	duplex_nup_value = duplex_nup_value ..'0'
	-- 3 mojime
	if( '9pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'1'
	elseif( '16pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'2'
	else
		duplex_nup_value = duplex_nup_value ..'0'
	end
	-- 4 mojime
	if( '2pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'1'
	elseif( '4pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'2'
	elseif( '6pagespersheet' == settings_table.layout ) then
		duplex_nup_value = duplex_nup_value ..'4'
	else
		duplex_nup_value = duplex_nup_value ..'0'
	end
	-- 5 mojime
	duplex_nup_value = duplex_nup_value ..'0'
	-- 6 mojime
	if( (settings_table.booklet)	and
		'off' ~= settings_table.booklet	) then
		duplex_nup_value = duplex_nup_value ..'8'
	elseif(	(settings_table.layout)		and
			'1pagepersheet' ~= settings_table.layout	) then
		if( (settings_table.duplex)	and
			'off' ~= settings_table.duplex ) then
			duplex_nup_value = duplex_nup_value ..'2'
		else
			duplex_nup_value = duplex_nup_value ..'1'
		end
	else
		duplex_nup_value = duplex_nup_value ..'0'
	end
	-- 7 mojime
	duplex_nup_value = duplex_nup_value ..'0'
	-- 8 mojime
	if(	'1pagepersheet' == settings_table.layout 	and
		(settings_table.duplex)						and
		'off' ~= settings_table.duplex					) then
		duplex_nup_value = duplex_nup_value..'1'
	else
		duplex_nup_value = duplex_nup_value..'0'
	end

	return 'DUPLEX_NUP='..duplex_nup_value
end
-----------------------------------------------------------------------------
-- ELocal Function : Generate OTHER in PRINTINFO
-----------------------------------------------------------------------------
function generate_other_printinfo(settings_table)
	local othe_value = 0

	if(	( 'on' == settings_table.pwmk_print )	or
		( 'on' == settings_table.pow_print )	or
		( nil ~= settings_table.phf_print	and
		  'off' ~= settings_table.phf_print )	) then
		othe_value = othe_value + 8
	end

	if(	(	nil ~= settings_table.booklet	and
			'off' ~= settings_table.booklet 	)	or
		(	nil ~= settings_table.poster	and
			'off' ~= settings_table.poster 		)		) then
		othe_value = othe_value + 4
	end

	if(	'on' == settings_table.frontcover 	or
		'on' == settings_table.backcover 	or
		'on' == settings_table.slipsheet 	or
		'on' == settings_table.chapterpages 	) then
		othe_value = othe_value + 2
	end

	return string.format('OTHER=0000000%X', othe_value)

end
-----------------------------------------------------------------------------
-- ELocal Function : Generate PRINTINFO
-----------------------------------------------------------------------------
function generate_printinfo(settings_table)
	local printinfo = '@PJL SET PRINTINFO = "'
	printinfo = printinfo..generate_duplex_nup_printinfo(settings_table)..'/'
	printinfo = printinfo..generate_scaling_printinfo(settings_table)..'/'
	printinfo = printinfo..generate_other_printinfo(settings_table)..'"\r\n'

	return printinfo
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate PROOFJOB
-----------------------------------------------------------------------------
function generate_jobtype(settings_table)
	return ({
		sampleprint				= '@PJL PROOFJOB\r\n',
		lockedprint				= '@PJL SECUREJOB\r\n',
		holdprint				= '@PJL PAUSEDJOB\r\n',
		storedprint				= '@PJL PAUSEDJOB\r\n',
		storedprintshared		= '@PJL PAUSEDJOB\r\n',
		documentserverrgb		= '@PJL REUSABLESTOREDJOB\r\n'
	})[settings_table.jobtype]
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate DATE
-----------------------------------------------------------------------------
function generate_date()
	local year	= os.date('%Y')
	local month	= os.date('%m')
	local day	= os.date('%d')

	local date = '2000/01/01'
	if	tonumber( year )	>= 1970	and
		tonumber( month )	>= 1	and
		tonumber( day )		>= 1	then
		date = string.format( '%s/%s/%s', year, month, day )
	end

	return '@PJL SET DATE = "' .. date  .. '"\r\n'
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate TIME
-----------------------------------------------------------------------------
function generate_time()
	local hour		= os.date('%H')
	local minute	= os.date('%M')
	local second	= os.date('%S')

	local time = '00:00:00'
	if	tonumber( hour )	>= 0	and
		tonumber( minute )	>= 0	and
		tonumber( second )	>= 0	then
		time = string.format( '%s:%s:%s', hour, minute, second )
	end

	return '@PJL SET TIME = "' .. time .. '"\r\n'
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate DATE&TIME
-----------------------------------------------------------------------------
function generate_date_and_time()
	local year		= os.date('%Y')
	local month		= os.date('%m')
	local hour		= os.date('%H')
	local day		= os.date('%d')
	local minute	= os.date('%M')

	local date = '2000/01/01 00:00'
	if	tonumber( year )	>= 1970	and
		tonumber( month )	>= 1	and
		tonumber( day )		>= 1	and
		tonumber( hour )	>= 0	and
		tonumber( minute )	>= 0	then
		date = string.format( '%s/%s/%s %s:%s', year, month, day, hour, minute )
	end

	return date
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate DRIVERKINDINFO
-----------------------------------------------------------------------------
function generate_driverkindinfo(settings_table)
	return ( nil ~= settings_table.geneinneritem_driver and "generic" == settings_table.geneinneritem_driver ) and '@PJL SET DRIVERKINDINFO = PCL6UDGENERIC\r\n' or '@PJL SET DRIVERKINDINFO = PCL6UD\r\n'
end

-----------------------------------------------------------------------------
-- ELocal Function : Generate CustomMediaType
-----------------------------------------------------------------------------
function generate_custommediatype(settings_table)
		
	local command_tbl = {}
	
	if( "raw" == settings_table.jh_custommedia_stringtype	) then
		for index = 1, 200 do
			if( nil ~= settings_table['jh_raw_custommediacharset' .. index] ) then
				table.insert( command_tbl, string.format('@PJL SET MEDIA%sNAMECHARSET = %s\r\n', index, settings_table['jh_raw_custommediacharset' .. index]) )
			end
			
			if( nil ~= settings_table['jh_raw_custommedianame' .. index] ) then
				table.insert( command_tbl, string.format('@PJL SET MEDIA%sNAME = "%s"\r\n', index, settings_table['jh_raw_custommedianame' .. index]) )
			end
		end
	else
		for index = 1, 200 do
			local charset_val = settings_table.jh_charset and settings_table.jh_charset or settings_table['jh_custommediacharset' .. index]
			
			if( 'on' == settings_table.gb18030support	and
				'8' == charset_val							) then
				charset_val = '33'
			elseif(not isMember( charset_val, { '1', '2', '3', '4', '5', '6', '8', '9', '11', '12', '13' } )) then
				charset_val = '2'
			end
			
			if( nil ~= settings_table['jh_custommediacharset' .. index] ) then
				table.insert( command_tbl, string.format('@PJL SET MEDIA%sNAMECHARSET = %s\r\n', index, charset_val) )
			end
			
			if( nil ~= settings_table['jh_custommedianame' .. index] ) then
				table.insert( command_tbl, string.format('@PJL SET MEDIA%sNAME = "%s"\r\n', index, get_trimmed_command_string( settings_table['jh_custommedianame' .. index], charset_val == '33' and 'utf8' or 'osdependent', 227 )) )
			end
		end
	end
	
	return table.concat(command_tbl)
end


-----------------------------------------------------------------------------
-- Export Function : Generate Command
-----------------------------------------------------------------------------
function generate_command(settings_table)

  local converter = TableConverter(convert_table)
  local job_commands = converter:run(settings_table)

  local commands = {}

  table.insert( commands, string.char(0x1b) .. '%-12345X')
  table.insert( commands, generate_begin_jobname(settings_table))
  table.insert( commands, generate_driverkindinfo(settings_table) )
  table.insert(	commands, generate_printinfo(settings_table) )
  table.insert(	commands, generate_pcl_comment(settings_table) )
  table.insert(	commands, generate_pcl_comment_other(settings_table) )
  table.insert(	commands, generate_jobtype(settings_table) )
  table.insert( commands, generate_date() )
  table.insert( commands, generate_time() )
  table.insert( commands, generate_custommediatype(settings_table) )
  table.insert( commands, generate_job_additional_info_userdomain_name(settings_table) )
  table.insert( commands, generate_job_additional_info_process_name(settings_table) )

  for item, value in pairs(job_commands) do
	table.insert(commands, string.format('@PJL SET %s = %s\r\n', item, value))
  end

  local loader = ExtensionLoader(_MODULE, 'generate_command' )
  for name, func in loader:extensions() do
     func(settings_table, commands)
  end

  table.insert( commands, '@PJL ENTER LANGUAGE = PCLXL\r\n' )

  return table.concat(commands)
end

return true
