Rice.require 'Rice.CapsDB'
Rice.require 'Rice.PaperDB'
Rice.require 'Rice.PrintTicket'
Rice.require 'TableConverter'

--------------------------------------------------------------------
-- Global Root Item
--------------------------------------------------------------------
printticket_xpath		= '/psf:PrintTicket'
printcapabilities_xpath	= '/psf:PrintCapabilities'
ptpc_version 			= '1'

xmlns_table =
{
	psf 	= 'http://schemas.microsoft.com/windows/2003/08/printing/printschemaframework',
	xsi 	= 'http://www.w3.org/2001/XMLSchema-instance',
	xsd		= 'http://www.w3.org/2001/XMLSchema',
	psk 	= 'http://schemas.microsoft.com/windows/2003/08/printing/printschemakeywords',
	ns0000	= 'http://schemas.ricoh.com/2007/printing/keywords'
}

--------------------------------------------------------------------
-- Local Function
--------------------------------------------------------------------
-- Get prefix conforming to specified uri from ptpc
local function get_prefix_from_ptpc( ptpc, uri )
	return ptpc.get_prefix(uri)
end

-- Get uri conforming to specified prefix from ptpc
local function get_uri_from_ptpc( ptpc, prefix )
	return ptpc.get_uri(prefix)
end

-- Get prefix conforming to specified uri from xmlns_table
local function get_prefix_from_table( uri )
	for i in pairs(xmlns_table) do
		if uri == xmlns_table[i] then
			return i
		end
	end
end

-- Get URI conforming to specified prefix from xmlns_table
local function get_uri_from_table( prefix )
	for i in pairs(xmlns_table) do
		if prefix == i then
			return xmlns_table[i]
		end
	end
end

local function generate_text( pt, xpath, namespace, text)
	return pt.set_text(	xpath,
						text,
						namespace )
end

local function generate_attribute( pt, xpath, namespace, tbl )

	local pt_result = false

	if( nil == pt or nil == xpath or nil == namespace or nil == tbl ) then
		return false
	end

	local isMultiAttribute = ( function()
		for i in pairs(tbl) do
			if( 'key' == i or 'value' == i ) then
				return false
			end
		end
		return true
	end)()

	if( isMultiAttribute ) then
		for i in pairs(tbl) do
			if( false == generate_attribute(tbl[i]) ) then
				return false
			end
		end
		pt_result = true
	else
		local attribute_namespace = ( function( key )
			local beginpos,endpos = string.find( key,':' )
			if( nil ~= beginpos and nil ~= endpos and endpos > 1 ) then
				return xmlns_table[string.sub( key, 1, endpos - 1 )]
			else
				return nil
			end
		end)(tbl.key)

		pt_result = pt.set_attribute(	xpath,
										tbl.key,
										tbl.value,
										attribute_namespace )
	end

	return pt_result
end

local function generate_tag( pt, xpath, tbl )
	if( nil == pt or nil == xpath or nil ==  tbl) then
		return false
	end

	local pt_result = false

	local isMultiTag = ( function()
		local count = 0
		for i in pairs(tbl) do
			if( 'number' ~= type(i) ) then
				return false
			end
			count = count + 1
		end
		return ( count == #tbl )
	end)()

	if( isMultiTag ) then
		for i, v in ipairs(tbl) do
			if( v and v.tag ) then
				if( not generate_tag(pt, xpath, v) ) then
					return false
				end
			end
		end
		pt_result = true
	else
		if( nil == tbl.tag ) then
			return true
		end
		local namespace = (function( tag )
			local beginpos,endpos = string.find( tag,':' )
			if( nil ~= beginpos and nil ~= endpos and endpos > 1 ) then
				return xmlns_table[string.sub( tag, 1, endpos - 1 )]
			else
				return nil
			end
		end)(tbl.tag)

		if( nil ~= tbl.tag ) then
			pt_result = pt.set_element(	xpath,
										tbl.tag,
										namespace )
		end
		if( true == pt_result and nil ~= tbl.attribute ) then
			pt_result = generate_attribute( pt, xpath..'/'..tbl.tag..'[last()]',
											namespace,
											tbl.attribute )
		end
		if( true == pt_result and nil ~= tbl.text ) then
			pt_result = generate_text( 	pt, xpath..'/'..tbl.tag..'[last()]',
										namespace,
										tbl.text)
		end


		if( true == pt_result and nil ~= tbl.children ) then
			pt_result = generate_tag(	pt, xpath..'/'..tbl.tag..'[last()]',
										tbl.children )
		end

	end
	return pt_result
end

local function convert_to_value( item_table, value )

	local result_value = nil

	if( "table" == type(value) ) then
		--for input TableConverter
		local TableConverterTable = {}
		for i, v in pairs(value) do
			TableConverterTable[i] =
			{
				item	=	'item',
				values	=	v,
				default	=	value.default
			}
		end

		local converter 		= TableConverter(TableConverterTable)
		local converted_table 	= converter:run(item_table)
		if( nil ~= converted_table and converted_table.item) then
			result_value = converted_table.item
		end

	end
	return result_value
end

--------------------------------------------------------------------
-- PTPC COMMON FUNCTION
--------------------------------------------------------------------
function create_option_info(cdb, paper_db)
	local option_info_result = {}

	if cdb and paper_db then
		local resolution = cdb.get_data_item_list('resolution')

		local paper_item = cdb.get_detail('documentsize')

		local paper_options = {}
		if paper_item then
			paper_options = paper_item['elements']
		end

		local printablearea = cdb.get('printablearea')
		if nil == printablearea then
			printablearea = ''
		end
		local edgetoedgeprint = cdb.get('edgetoedgeprint')
		if nil == edgetoedgeprint then
			edgetoedgeprint = ''
		end
			
		local paper_info_table = {}
		if resolution and resolution[1] then
			if 'on' == edgetoedgeprint then
				for i in pairs(paper_options) do
					local paper_info = {}
	
					local paper_info_mm = paper_db.get('', i, 'mm', tonumber(resolution[1]))
					paper_info_mm['page_width'] = paper_info_mm['width']
					paper_info_mm['page_height'] = paper_info_mm['height']
					paper_info_mm['offset_x'] = 0
					paper_info_mm['offset_y'] = 0
					
					paper_info['mm'] = paper_info_mm
	
					local paper_info_inch = paper_db.get('', i, 'inch', tonumber(resolution[1]))
					paper_info_inch['page_width'] = paper_info_inch['width']
					paper_info_inch['page_height'] = paper_info_inch['height']
					paper_info_inch['offset_x'] = 0
					paper_info_inch['offset_y'] = 0
					
					paper_info['inch'] = paper_info_inch
	
					local paper_info_pixel = paper_db.get('', i, 'pixel', tonumber(resolution[1]))
					paper_info_pixel['page_width'] = paper_info_pixel['width']
					paper_info_pixel['page_height'] = paper_info_pixel['height']
					paper_info_pixel['offset_x'] = 0
					paper_info_pixel['offset_y'] = 0
					
					paper_info['pixel'] = paper_info_pixel
	
					paper_info_table[i] = paper_info
				end
			else
				for i in pairs(paper_options) do
					local paper_info = {}
	
					local paper_info_mm = paper_db.get(printablearea, i, 'mm', tonumber(resolution[1]))
					paper_info['mm'] = paper_info_mm
					
					local paper_info_inch = paper_db.get(printablearea, i, 'inch', tonumber(resolution[1]))
					paper_info['inch'] = paper_info_inch
	
					local paper_info_pixel = paper_db.get(printablearea, i, 'pixel', tonumber(resolution[1]))
					paper_info['pixel'] = paper_info_pixel
					
					paper_info_table[i] = paper_info
				end
			end

			option_info_result['current_resolution'] = resolution[1]
			option_info_result['paper_info'] = paper_info_table
		end
	end

	return option_info_result
end

function get_duplex_version(cdb)
	local element = {}
	local duplex_version = 'duplexversion2'
	
	if ( cdb ) then
		element = cdb.get_pickone_element_list('duplex')
		if ( element and element[1] ) then
			for i, item in pairs(element) do
				if('longedgeleftortop' == item ) then
					duplex_version = 'duplexversion2' do break end
				elseif( 'opentoleft' == item ) then
					duplex_version = 'duplexversion1' do break end
				end
			end
		end
	end
	return duplex_version
end

-- Replace prefix
-- If qname doesn't have prefix or conversion failed then return specified qname without modification
function solve_local_prefix(ptpc, qname)

	local i,j= string.find(qname, '^.-:')
	if i ~= nil then
		local org_prefix = string.sub(qname, i, j-1)

		local uri = get_uri_from_table(org_prefix)

		local actual_prefix = get_prefix_from_ptpc(ptpc,uri)
		if actual_prefix == nil then
			return qname
		else
			return string.gsub(qname, org_prefix, actual_prefix, 1)
		end
	else
		return qname
	end
end

-- Replace prefix
-- If qname doesn't have prefix or conversion failed then return specified qname without modification
function solve_actual_prefix(ptpc, qname)

	local i,j= string.find(qname, '^.-:')
	if i ~= nil then
		local org_prefix = string.sub(qname, i, j-1)

		local uri = get_uri_from_ptpc(ptpc,org_prefix)

		local prefix = get_prefix_from_table(uri)

		if prefix == nil then
			return qname
		else
			return string.gsub(qname, org_prefix, prefix, 1)
		end

	else
		return qname
	end

end

function set_namespace(ptpc_Obj)
	local result = false

	if ptpc_Obj then
		local namespace_string = ""
		for i in pairs(xmlns_table) do

			local prefix = get_prefix_from_ptpc(ptpc_Obj, xmlns_table[i])
			if nil == prefix or '' == prefix then
				prefix = i
			end
			namespace_string = namespace_string.."xmlns:"..prefix.."='"..xmlns_table[i].."' "
		end
		ptpc_Obj.set_namespace(namespace_string)

		result = true
	end

	return result
end

function set_root_attribute(ptpc_Obj,xpath)
	local result = false
	
	if ptpc_Obj and xpath then
		for i in pairs(xmlns_table) do
			ptpc_Obj.set_attribute(	xpath,
									'xmlns:'..i,
									xmlns_table[i],
									xmlns_table.psf )

		end
		ptpc_Obj.set_attribute(	xpath,
								'version',
								ptpc_version,
								nil )
		result = true
	end

	return result
end

function get_all_items( cdb )
	local item_table = {}

	if( nil ~= cdb ) then
		local item_list = cdb.features()
		if( nil ~= item_list ) then
			for i, v in ipairs(item_list) do
				local cdb_get_result = cdb.get(v)
				item_table[v] = cdb_get_result and cdb_get_result or ''
			end
		end
	end

	return item_table
end

function convert_ptpc_table( item_table, ptpc_table, option_info)
	local result_ptpc_table = {}
	for i in pairs(ptpc_table) do
		if( 'string' == type(ptpc_table[i]) or 'nil' == type(ptpc_table[i])) then
			result_ptpc_table[i] = ptpc_table[i]
		elseif( 'function' == type(ptpc_table[i]) ) then
			result_ptpc_table[i] = ptpc_table[i](item_table, option_info)
		elseif( 'table'	== type( ptpc_table[i]) ) then
			if( 'tag' == i or 'key' == i or 'value' == i or 'text' == i) then
				result_ptpc_table[i] = convert_to_value(item_table,ptpc_table[i])
			else
				result_ptpc_table[i] = convert_ptpc_table( item_table, ptpc_table[i], option_info)
			end
		end
	end
	return result_ptpc_table
end

-- Devmode to PrintTicket Code
function generate_printtickets( pt, pt_tbl )
	if( nil == pt or nil ==  pt_tbl ) then
		return false
	end

	for i, v in ipairs(pt_tbl) do
		if( nil ~= v and false == generate_tag( pt, printticket_xpath, v) ) then
			return false
		end
	end
	return true
end

-- PrintCapabilies Code
function generate_printcapabilities( pc, item_table)
	if( nil == pc or nil ==  item_table ) then
		return false
	end

	for i, v in ipairs(item_table) do
		if( nil ~= v and false == generate_tag( pc, printcapabilities_xpath, v) ) then
			return false
		end
	end
	return true
end
